# SPDX-License-Identifier: MIT
import io, sys, traceback, struct, array, bisect, os, plistlib, signal, runpy
from construct import *

from ..asm import ARMAsm
from ..tgtypes import *
from ..proxy import IODEV, START, EVENT, EXC, EXC_RET, ExcInfo
from ..utils import *
from ..sysreg import *
from ..macho import MachO
from ..adt import load_adt
from .. import xnutools, shell

from .gdbserver import *
from .types import *
from .virtutils import *
from .virtio import *

__all__ = ["HV"]

class HV(Reloadable):
    PAC_MASK = 0xfffff00000000000

    PTE_VALID               = 1 << 0

    PTE_MEMATTR_UNCHANGED   = 0b1111 << 2
    PTE_S2AP_RW             = 0b11 << 6
    PTE_SH_NS               = 0b11 << 8
    PTE_ACCESS              = 1 << 10
    PTE_ATTRIBUTES          = PTE_ACCESS | PTE_SH_NS | PTE_S2AP_RW | PTE_MEMATTR_UNCHANGED

    SPTE_TRACE_READ         = 1 << 63
    SPTE_TRACE_WRITE        = 1 << 62
    SPTE_TRACE_UNBUF        = 1 << 61
    SPTE_MAP                = 0 << 50
    SPTE_HOOK               = 1 << 50
    SPTE_PROXY_HOOK_R       = 2 << 50
    SPTE_PROXY_HOOK_W       = 3 << 50
    SPTE_PROXY_HOOK_RW      = 4 << 50

    MSR_REDIRECTS = {
        SCTLR_EL1: SCTLR_EL12,
        TTBR0_EL1: TTBR0_EL12,
        TTBR1_EL1: TTBR1_EL12,
        TCR_EL1: TCR_EL12,
        ESR_EL1: ESR_EL12,
        FAR_EL1: FAR_EL12,
        AFSR0_EL1: AFSR0_EL12,
        AFSR1_EL1: AFSR1_EL12,
        MAIR_EL1: MAIR_EL12,
        AMAIR_EL1: AMAIR_EL12,
        CONTEXTIDR_EL1: CONTEXTIDR_EL12,
        # ACTLR_EL1: ACTLR_EL12, # Handled in hv_exc.c, depends on CPU version
        AMX_CONFIG_EL1: AMX_CONFIG_EL12,
        SPRR_CONFIG_EL1: SPRR_CONFIG_EL12,
        SPRR_PPERM_EL1: SPRR_PPERM_EL12,
        SPRR_UPERM_EL0: SPRR_UPERM_EL02,
        SPRR_AMRANGE_EL1: SPRR_AMRANGE_EL12,
        SPRR_UMPRR_EL1: SPRR_UMPRR_EL12,
        APCTL_EL1: APCTL_EL12,
        APSTS_EL1: APSTS_EL12,
        KERNKEYLO_EL1: KERNKEYLO_EL12,
        KERNKEYHI_EL1: KERNKEYHI_EL12,
        GXF_CONFIG_EL1: GXF_CONFIG_EL12,
        GXF_PABENTRY_EL1: GXF_PABENTRY_EL12,
        GXF_ENTRY_EL1: GXF_ENTRY_EL12,
        VBAR_GL1: VBAR_GL12,
        SPSR_GL1: SPSR_GL12,
        ASPSR_GL1: ASPSR_GL12,
        ESR_GL1: ESR_GL12,
        ELR_GL1: ELR_GL12,
    }

    AIC_EVT_TYPE_HW = 1
    IRQTRACE_IRQ = 1

    def __init__(self, iface, proxy, utils):
        self.iface = iface
        self.p = proxy
        self.u = utils
        self.pac_mask = self.PAC_MASK
        self.user_pac_mask = self.PAC_MASK
        self.vbar_el1 = None
        self.want_vbar = None
        self.vectors = [None]
        self._bps = [None, None, None, None, None]
        self._bp_hooks = dict()
        self._wps = [None, None, None, None]
        self._wpcs = [0, 0, 0, 0]
        self.sym_offset = 0
        self.symbols = []
        self.symbol_dict = {}
        self.sysreg = {}
        self.novm = False
        self._in_handler = False
        self._sigint_pending = False
        self._in_shell = False
        self._gdbserver = None
        self.vm_hooks = [None]
        self.interrupt_map = {}
        self.mmio_maps = DictRangeMap()
        self.dirty_maps = BoolRangeMap()
        self.tracer_caches = {}
        self.shell_locals = {}
        self.xnu_mode = False
        self._update_shell_locals()
        self.wdt_cpu = None
        self.smp = True
        self.hook_exceptions = False
        self.started_cpus = {}
        self.started = False
        self.ctx = None
        self.hvcall_handlers = {}
        self.switching_context = False
        self.show_timestamps = False
        self.virtio_devs = {}

    def _reloadme(self):
        super()._reloadme()
        self._update_shell_locals()

    def _update_shell_locals(self):
        self.shell_locals.update({
            "hv": self,
            "iface": self.iface,
            "p": self.p,
            "u": self.u,
            "trace": trace,
            "TraceMode": TraceMode,
        })

        for attr in dir(self):
            a = getattr(self, attr)
            if callable(a):
                self.shell_locals[attr] = getattr(self, attr)

        self.shell_locals["ctx"] = self.context

    def log(self, s, *args, show_cpu=True, **kwargs):
        if self.ctx is not None and show_cpu:
            ts=""
            if self.show_timestamps:
                ts = f"[{self.u.mrs(CNTPCT_EL0):#x}]"
            print(ts+f"[cpu{self.ctx.cpu_id}] " + s, *args, **kwargs)
            if self.print_tracer.log_file:
                print(f"# {ts}[cpu{self.ctx.cpu_id}] " + s, *args, file=self.print_tracer.log_file, **kwargs)
        else:
            print(s, *args, **kwargs)
            if self.print_tracer.log_file:
                print("# " + s, *args, file=self.print_tracer.log_file, **kwargs)

    def unmap(self, ipa, size):
        assert self.p.hv_map(ipa, 0, size, 0) >= 0

    def map_hw(self, ipa, pa, size):
        '''map IPA (Intermediate Physical Address) to actual PA'''
        #print(f"map_hw {ipa:#x} -> {pa:#x} [{size:#x}]")
        if (ipa & 0x3fff) != (pa & 0x3fff):
            self.map_sw(ipa, pa, size)
            return

        ipa_p = align_up(ipa)
        if ipa_p != ipa:
            self.map_sw(ipa, pa, min(ipa_p - ipa, size))
            pa += ipa_p - ipa
            size -= ipa_p - ipa

        if size <= 0:
            return

        size_p = align_down(size)
        if size_p > 0:
            #print(f"map_hw real {ipa_p:#x} -> {pa:#x} [{size_p:#x}]")
            assert self.p.hv_map(ipa_p, pa | self.PTE_ATTRIBUTES | self.PTE_VALID, size_p, 1) >= 0

        if size_p != size:
            self.map_sw(ipa_p + size_p, pa + size_p, size - size_p)

    def map_sw(self, ipa, pa, size):
        #print(f"map_sw {ipa:#x} -> {pa:#x} [{size:#x}]")
        assert self.p.hv_map(ipa, pa | self.SPTE_MAP, size, 1) >= 0

    def map_hook(self, ipa, size, read=None, write=None, **kwargs):
        index = len(self.vm_hooks)
        self.vm_hooks.append((read, write, ipa, kwargs))
        self.map_hook_idx(ipa, size, index, read is not None, write is not None)

    def map_hook_idx(self, ipa, size, index, read=False, write=False, flags=0):
        if read:
            if write:
                t = self.SPTE_PROXY_HOOK_RW
            else:
                t = self.SPTE_PROXY_HOOK_R
        elif write:
            t = self.SPTE_PROXY_HOOK_W
        else:
            assert False

        assert self.p.hv_map(ipa, (index << 2) | flags | t, size, 0) >= 0

    def readmem(self, va, size):
        '''read from virtual memory'''
        with io.BytesIO() as buffer:
            while size > 0:
                pa = self.p.hv_translate(va, False, False)
                if pa == 0:
                    break

                size_in_page = 4096 - (va % 4096)
                if size < size_in_page:
                    buffer.write(self.iface.readmem(pa, size))
                    break

                buffer.write(self.iface.readmem(pa, size_in_page))
                va += size_in_page
                size -= size_in_page

            return buffer.getvalue()

    def writemem(self, va, data):
        '''write to virtual memory'''
        written = 0
        while written < len(data):
            pa = self.p.hv_translate(va, False, True)
            if pa == 0:
                break

            size_in_page = 4096 - (va % 4096)
            if len(data) - written < size_in_page:
                self.iface.writemem(pa, data[written:])
                written = len(data)
                break

            self.iface.writemem(pa, data[written:written + size_in_page])
            va += size_in_page
            written += size_in_page

        return written

    def trace_irq(self, device, num, count, flags):
        for n in range(num, num + count):
            if flags & self.IRQTRACE_IRQ:
                self.interrupt_map[n] = device
            else:
                self.interrupt_map.pop(n, None)

        start, size = self.adt["/arm-io/aic"].get_reg(0)
        zone = irange(start, size)
        if len(self.interrupt_map):
            self.add_tracer(zone, "AIC_IRQ", TraceMode.RESERVED)
        else:
            self.del_tracer(zone, "AIC_IRQ")

        assert self.p.hv_trace_irq(self.AIC_EVT_TYPE_HW, num, count, flags) > 0

    def add_tracer(self, zone, ident, mode=TraceMode.ASYNC, read=None, write=None, **kwargs):
        assert mode in (TraceMode.RESERVED, TraceMode.OFF, TraceMode.BYPASS) or read or write
        self.mmio_maps[zone, ident] = (mode, ident, read, write, kwargs)
        self.dirty_maps.set(zone)

    def del_tracer(self, zone, ident):
        del self.mmio_maps[zone, ident]
        self.dirty_maps.set(zone)

    def clear_tracers(self, ident):
        for r, v in self.mmio_maps.items():
            if ident in v:
                v.pop(ident)
                self.dirty_maps.set(r)

    def trace_device(self, path, mode=TraceMode.ASYNC, ranges=None):
        node = self.adt[path]
        for index in range(len(node.reg)):
            if ranges is not None and index not in ranges:
                continue
            addr, size = node.get_reg(index)
            self.trace_range(irange(addr, size), mode)

    def trace_range(self, zone, mode=TraceMode.ASYNC, read=True, write=True, name=None):
        if mode is True:
            mode = TraceMode.ASYNC
        if mode and mode != TraceMode.OFF:
            self.add_tracer(zone, "PrintTracer", mode,
                            self.print_tracer.event_mmio if read else None,
                            self.print_tracer.event_mmio if write else None,
                            start=zone.start,
                            name=name)
        else:
            self.del_tracer(zone, "PrintTracer")

    def pt_update(self):
        if not self.dirty_maps:
            return

        self.dirty_maps.compact()
        self.mmio_maps.compact()

        top = 0

        for zone in self.dirty_maps:
            if zone.stop <= top:
                continue
            top = max(top, zone.start)

            for mzone, maps in self.mmio_maps.overlaps(zone):
                if mzone.stop <= top:
                    continue
                if top < mzone.start:
                    self.unmap(top, mzone.start - top)
                    self.log(f"PT[{top:09x}:{mzone.start:09x}] -> *UNMAPPED*")

                top = mzone.stop
                if not maps:
                    continue
                maps = sorted(maps.values(), reverse=True)
                mode, ident, read, write, kwargs = maps[0]

                need_read = any(m[2] for m in maps)
                need_write = any(m[3] for m in maps)

                if mode == TraceMode.RESERVED:
                    self.log(f"PT[{mzone.start:09x}:{mzone.stop:09x}] -> RESERVED {ident}")
                    continue
                elif mode in (TraceMode.HOOK, TraceMode.SYNC):
                    self.map_hook_idx(mzone.start, mzone.stop - mzone.start, 0,
                                      need_read, need_write)
                    if mode == TraceMode.HOOK:
                        for m2, i2, r2, w2, k2 in maps[1:]:
                            if m2 == TraceMode.HOOK:
                                self.log(f"!! Conflict: HOOK {i2}")
                elif mode == TraceMode.WSYNC:
                    flags = self.SPTE_TRACE_READ if need_read else 0
                    self.map_hook_idx(mzone.start, mzone.stop - mzone.start, 0,
                                      False, need_write, flags=flags)
                elif mode in (TraceMode.UNBUF, TraceMode.ASYNC, TraceMode.BYPASS):
                    pa = mzone.start
                    if mode == TraceMode.UNBUF:
                        pa |= self.SPTE_TRACE_UNBUF
                    if need_read:
                        pa |= self.SPTE_TRACE_READ
                    if need_write:
                        pa |= self.SPTE_TRACE_WRITE
                    self.map_sw(mzone.start, pa, mzone.stop - mzone.start)
                elif mode == TraceMode.OFF:
                    self.map_hw(mzone.start, mzone.start, mzone.stop - mzone.start)
                    self.log(f"PT[{mzone.start:09x}:{mzone.stop:09x}] -> HW:{ident}")
                    continue

                rest = [m[1] for m in maps[1:] if m[0] != TraceMode.OFF]
                if rest:
                    rest = " (+ " + ", ".join(rest) + ")"
                else:
                    rest = ""

                self.log(f"PT[{mzone.start:09x}:{mzone.stop:09x}] -> {mode.name}.{'R' if read else ''}{'W' if read else ''} {ident}{rest}")

            if top < zone.stop:
                self.unmap(top, zone.stop - top)
                self.log(f"PT[{top:09x}:{zone.stop:09x}] -> *UNMAPPED*")

        self.u.inst(0xd50c83df) # tlbi vmalls12e1is
        self.dirty_maps.clear()

    def shellwrap(self, func, description, update=None, needs_ret=False):

        while True:
            try:
                return func()
            except Exception:
                print(f"Exception in {description}")
                traceback.print_exc()

            if not self.ctx:
                print("Running in asynchronous context. Target operations are not available.")

            def do_exit(i):
                raise shell.ExitConsole(i)

            self.shell_locals["skip"] = lambda: do_exit(1)
            self.shell_locals["cont"] = lambda: do_exit(0)
            ret = self.run_shell("Entering debug shell", "Returning to tracer")
            self.shell_locals["skip"] = self.skip
            self.shell_locals["cont"] = self.cont

            if self.ctx:
                self.cpu() # Return to the original CPU to avoid confusing things

            if ret == 1:
                if needs_ret:
                    print("Cannot skip, return value required.")
                else:
                    return

            if update:
                update()

    def run_shell(self, entry_msg="Entering shell", exit_msg="Continuing"):
        def handle_sigusr1(signal, stack):
            raise shell.ExitConsole(EXC_RET.HANDLED)

        def handle_sigusr2(signal, stack):
            raise shell.ExitConsole(EXC_RET.EXIT_GUEST)

        default_sigusr1 = signal.signal(signal.SIGUSR1, handle_sigusr1)
        try:
            default_sigusr2 = signal.signal(signal.SIGUSR2, handle_sigusr2)
            try:
                self._in_shell = True
                try:
                    if not self._gdbserver is None:
                        self._gdbserver.notify_in_shell()
                    return shell.run_shell(self.shell_locals, entry_msg, exit_msg)
                finally:
                    self._in_shell = False
            finally:
                signal.signal(signal.SIGUSR2, default_sigusr2)
        finally:
            signal.signal(signal.SIGUSR1, default_sigusr1)

    @property
    def in_shell(self):
        return self._in_shell

    def gdbserver(self, address="/tmp/.m1n1-unix", log=None):
        '''activate gdbserver'''
        if not self._gdbserver is None:
            raise Exception("gdbserver is already running")

        self._gdbserver = GDBServer(self, address, log)
        self._gdbserver.activate()

    def shutdown_gdbserver(self):
        '''shutdown gdbserver'''
        self._gdbserver.shutdown()
        self._gdbserver = None

    def handle_mmiotrace(self, data):
        evt = EvtMMIOTrace.parse(data)

        def do_update():
            nonlocal mode, ident, read, write, kwargs
            read = lambda *args, **kwargs: None
            write = lambda *args, **kwargs: None

            m = self.mmio_maps[evt.addr].get(ident, None)
            if not m:
                return

            mode, ident, read_, write_, kwargs = m
            read = read_ or read
            write = write_ or write

        maps = sorted(self.mmio_maps[evt.addr].values(), reverse=True)
        for mode, ident, read, write, kwargs in maps:
            if mode > TraceMode.WSYNC or (evt.flags.WRITE and mode > TraceMode.UNBUF):
                print(f"ERROR: mmiotrace event but expected {mode.name} mapping")
                continue
            if mode == TraceMode.OFF:
                continue
            if evt.flags.WRITE:
                if write:
                    self.shellwrap(lambda: write(evt, **kwargs),
                                   f"Tracer {ident}:write ({mode.name})", update=do_update)
            else:
                if read:
                    self.shellwrap(lambda: read(evt, **kwargs),
                                   f"Tracer {ident}:read ({mode.name})", update=do_update)

    def handle_vm_hook_mapped(self, ctx, data):
        maps = sorted(self.mmio_maps[data.addr].values(), reverse=True)

        if not maps:
            raise Exception(f"VM hook without a mapping at {data.addr:#x}")

        def do_update():
            nonlocal mode, ident, read, write, kwargs
            read = lambda *args, **kwargs: None
            write = lambda *args, **kwargs: None

            m = self.mmio_maps[data.addr].get(ident, None)
            if not m:
                return

            mode, ident, read_, write_, kwargs = m
            read = read_ or read
            write = write_ or write

        mode, ident, read, write, kwargs = maps[0]

        first = 0

        val = data.data

        if mode not in (TraceMode.HOOK, TraceMode.SYNC, TraceMode.WSYNC):
            raise Exception(f"VM hook with unexpected mapping at {data.addr:#x}: {maps[0][0].name}")

        if not data.flags.WRITE:
            if mode == TraceMode.HOOK and not read:
                mode = TraceMode.SYNC
                first += 1

            if mode == TraceMode.HOOK:
                val = self.shellwrap(lambda: read(data.addr, 8 << data.flags.WIDTH, **kwargs),
                                     f"Tracer {ident}:read (HOOK)", update=do_update, needs_ret=True)

                if not isinstance(val, list) and not isinstance(val, tuple):
                    val = [val]
                first += 1
            elif mode == TraceMode.SYNC:
                try:
                    val = self.u.read(data.addr, 8 << data.flags.WIDTH)
                except:
                    self.log(f"MMIO read failed: {data.addr:#x} (w={data.flags.WIDTH})")
                    raise
                if not isinstance(val, list) and not isinstance(val, tuple):
                    val = [val]
            elif mode == TraceMode.WSYNC:
                raise Exception(f"VM hook with unexpected mapping at {data.addr:#x}: {maps[0][0].name}")

            for i in range(1 << max(0, data.flags.WIDTH - 3)):
                self.p.write64(ctx.data + 16 + 8 * i, val[i])

        elif mode == TraceMode.HOOK:
            first += 1

        flags = data.flags.copy()
        flags.CPU = self.ctx.cpu_id
        width = data.flags.WIDTH

        if width > 3:
            flags.WIDTH = 3
            flags.MULTI = 1

        for i in range(1 << max(0, width - 3)):
            evt = Container(
                flags = flags,
                reserved = 0,
                pc = ctx.elr,
                addr = data.addr + 8 * i,
                data = val[i]
            )

            for mode, ident, read, write, kwargs in maps[first:]:
                if flags.WRITE:
                    if write:
                        self.shellwrap(lambda: write(evt, **kwargs),
                                       f"Tracer {ident}:write ({mode.name})", update=do_update)
                else:
                    if read:
                        self.shellwrap(lambda: read(evt, **kwargs),
                                       f"Tracer {ident}:read ({mode.name})", update=do_update)

        if data.flags.WRITE:
            mode, ident, read, write, kwargs = maps[0]

            if data.flags.WIDTH <= 3:
                wval = val[0]
            else:
                wval = val

            if mode == TraceMode.HOOK and not write:
                mode = TraceMode.SYNC

            if mode == TraceMode.HOOK:
                self.shellwrap(lambda: write(data.addr, wval, 8 << data.flags.WIDTH, **kwargs),
                            f"Tracer {ident}:write (HOOK)", update=do_update)
            elif mode in (TraceMode.SYNC, TraceMode.WSYNC):
                try:
                    self.u.write(data.addr, wval, 8 << data.flags.WIDTH)
                except:
                    if data.flags.WIDTH > 3:
                        wval = wval[0]
                    self.log(f"MMIO write failed: {data.addr:#x} = {wval} (w={data.flags.WIDTH})")
                    raise

        return True

    def handle_vm_hook(self, ctx):
        data = self.iface.readstruct(ctx.data, VMProxyHookData)

        if data.id == 0:
            return self.handle_vm_hook_mapped(ctx, data)

        rfunc, wfunc, base, kwargs = self.vm_hooks[data.id]

        d = data.data
        if data.flags.WIDTH < 3:
            d = d[0]

        if data.flags.WRITE:
            wfunc(base, data.addr - base, d, 8 << data.flags.WIDTH, **kwargs)
        else:
            val = rfunc(base, data.addr - base, 8 << data.flags.WIDTH, **kwargs)
            if not isinstance(val, list) and not isinstance(val, tuple):
                val = [val]
            for i in range(1 << max(0, data.flags.WIDTH - 3)):
                self.p.write64(ctx.data + 16 + 8 * i, val[i])

        return True

    def handle_irqtrace(self, data):
        evt = EvtIRQTrace.parse(data)

        if evt.type == self.AIC_EVT_TYPE_HW and evt.flags & self.IRQTRACE_IRQ:
            dev = self.interrupt_map[int(evt.num)]
            print(f"IRQ: {dev}: {evt.num}")

    def addr(self, addr):
        unslid_addr = addr + self.sym_offset
        if self.xnu_mode and (addr < self.tba.virt_base or unslid_addr < self.macho.vmin):
            return f"0x{addr:x}"

        saddr, name = self.sym(addr)

        if name is None:
            return f"0x{addr:x} (0x{unslid_addr:x})"

        return f"0x{addr:x} ({name}+0x{unslid_addr - saddr:x})"

    def resolve_symbol(self, name):
        return self.symbol_dict[name] - self.sym_offset

    def sym(self, addr):
        unslid_addr = addr + self.sym_offset

        if self.xnu_mode and (addr < self.tba.virt_base or unslid_addr < self.macho.vmin):
            return None, None

        idx = bisect.bisect_left(self.symbols, (unslid_addr + 1, "")) - 1
        if idx < 0 or idx >= len(self.symbols):
            return None, None

        return self.symbols[idx]

    def get_sym(self, addr):
        a, name = self.sym(addr)
        if addr == a:
            return name
        else:
            return None

    def handle_msr(self, ctx, iss=None):
        if iss is None:
            iss = ctx.esr.ISS
        iss = ESR_ISS_MSR(iss)
        enc = iss.Op0, iss.Op1, iss.CRn, iss.CRm, iss.Op2

        name = sysreg_name(enc)

        skip = set()
        shadow = {
            #SPRR_CONFIG_EL1,
            #SPRR_PERM_EL0,
            #SPRR_PERM_EL1,
            VMSA_LOCK_EL1,
            #SPRR_UNK1_EL1,
            #SPRR_UNK2_EL1,
            MDSCR_EL1,
        }
        ro = {
            ACC_CFG_EL1,
            ACC_OVRD_EL1,
        }
        xlate = {
            DC_CIVAC,
        }
        for i in range(len(self._bps)):
            shadow.add(DBGBCRn_EL1(i))
            shadow.add(DBGBVRn_EL1(i))
        for i in range(len(self._wps)):
            shadow.add(DBGWCRn_EL1(i))
            shadow.add(DBGWVRn_EL1(i))

        value = 0
        if enc == CYC_OVRD_EL1 and iss.DIR == MSR_DIR.WRITE:
            if iss.Rt != 31:
                value = ctx.regs[iss.Rt]
            self.log(f"Skip: msr {name}, x{iss.Rt} = {value:x}")
            if value & 1:
                self.log("Guest is shutting down CPU")
                self.p.hv_exit_cpu()
                del self.started_cpus[self.ctx.cpu_id]
        elif enc in shadow:
            if iss.DIR == MSR_DIR.READ:
                value = self.sysreg[self.ctx.cpu_id].setdefault(enc, 0)
                self.log(f"Shadow: mrs x{iss.Rt}, {name} = {value:x}")
                if iss.Rt != 31:
                    ctx.regs[iss.Rt] = value
            else:
                if iss.Rt != 31:
                    value = ctx.regs[iss.Rt]
                self.log(f"Shadow: msr {name}, x{iss.Rt} = {value:x}")
                self.sysreg[self.ctx.cpu_id][enc] = value
        elif enc in skip or (enc in ro and iss.DIR == MSR_DIR.WRITE):
            if iss.DIR == MSR_DIR.READ:
                self.log(f"Skip: mrs x{iss.Rt}, {name} = 0")
                if iss.Rt != 31:
                    ctx.regs[iss.Rt] = 0
            else:
                if iss.Rt != 31:
                    value = ctx.regs[iss.Rt]
                self.log(f"Skip: msr {name}, x{iss.Rt} = {value:x}")
        else:
            if iss.DIR == MSR_DIR.READ:
                enc2 = self.MSR_REDIRECTS.get(enc, enc)
                value = self.u.mrs(enc2)
                self.log(f"Pass: mrs x{iss.Rt}, {name} = {value:x} ({sysreg_name(enc2)})")
                if iss.Rt != 31:
                    ctx.regs[iss.Rt] = value
            else:
                if iss.Rt != 31:
                    value = ctx.regs[iss.Rt]
                enc2 = self.MSR_REDIRECTS.get(enc, enc)
                sys.stdout.flush()
                if enc in xlate:
                    value = self.p.hv_translate(value, True, False)
                self.u.msr(enc2, value, call=self.p.gl2_call)
                self.log(f"Pass: msr {name}, x{iss.Rt} = {value:x} (OK) ({sysreg_name(enc2)})")

        ctx.elr += 4

        if self.hook_exceptions:
            self.patch_exception_handling()

        return True

    def handle_impdef(self, ctx):
        if ctx.esr.ISS == 0x20:
            return self.handle_msr(ctx, ctx.afsr1)

        code = struct.unpack("<I", self.iface.readmem(ctx.elr_phys, 4))
        c = ARMAsm(".inst " + ",".join(str(i) for i in code), ctx.elr_phys)
        insn = "; ".join(c.disassemble())

        self.log(f"IMPDEF exception on: {insn}")

        return False

    def handle_hvc(self, ctx):
        idx = ctx.esr.ISS
        if idx == 0:
            return False

        vector, target = self.vectors[idx]
        if target is None:
            self.log(f"EL1: Exception #{vector} with no target")
            target = 0
            ok = False
        else:
            ctx.elr = target
            ctx.elr_phys = self.p.hv_translate(target, False, False)
            ok = True

        if (vector & 3) == EXC.SYNC:
            spsr = SPSR(self.u.mrs(SPSR_EL12))
            esr = ESR(self.u.mrs(ESR_EL12))
            elr = self.u.mrs(ELR_EL12)
            elr_phys = self.p.hv_translate(elr, False, False)
            sp_el1 = self.u.mrs(SP_EL1)
            sp_el0 = self.u.mrs(SP_EL0)
            far = None
            if esr.EC == ESR_EC.DABORT or esr.EC == ESR_EC.IABORT:
                far = self.u.mrs(FAR_EL12)
                if self.sym(elr)[1] != "com.apple.kernel:_panic_trap_to_debugger":
                    self.log("Page fault")
                    return ok

            self.log(f"EL1: Exception #{vector} ({esr.EC!s}) to {self.addr(target)} from {spsr.M.name}")
            self.log(f"     ELR={self.addr(elr)} (0x{elr_phys:x})")
            self.log(f"     SP_EL1=0x{sp_el1:x} SP_EL0=0x{sp_el0:x}")
            if far is not None:
                self.log(f"     FAR={self.addr(far)}")
            if elr_phys:
                self.u.disassemble_at(elr_phys - 4 * 4, 9 * 4, elr - 4 * 4, elr, sym=self.get_sym)
            if self.sym(elr)[1] == "com.apple.kernel:_panic_trap_to_debugger":
                self.log("Panic! Trying to decode panic...")
                try:
                    self.decode_panic_call()
                except:
                    self.log("Error decoding panic.")
                try:
                    self.bt()
                except:
                    pass
                return False
            if esr.EC == ESR_EC.UNKNOWN:
                instr = self.p.read32(elr_phys)
                if instr == 0xe7ffdeff:
                    self.log("Debugger break! Trying to decode panic...")
                    try:
                        self.decode_dbg_panic()
                    except:
                        self.log("Error decoding panic.")
                    try:
                        self.bt()
                    except:
                        pass
                    return False
                return False
        else:
            elr = self.u.mrs(ELR_EL12)
            self.log(f"Guest: {str(EXC(vector & 3))} at {self.addr(elr)}")

        return ok

    def handle_step(self, ctx):
        # not sure why MDSCR_EL1.SS needs to be disabled here but otherwise
        # if also SPSR.SS=0 no instruction will be executed after eret
        # and instead a debug exception is generated again
        self.u.msr(MDSCR_EL1, MDSCR(MDE=1).value)

        # enable all breakpoints again
        for i, vaddr in enumerate(self._bps):
            if vaddr is None:
                continue
            self.u.msr(DBGBCRn_EL1(i), DBGBCR(E=1, PMC=0b11, BAS=0xf).value)

        # enable all watchpoints again
        for i, wpc in enumerate(self._wpcs):
            self.u.msr(DBGWCRn_EL1(i), wpc)

        return True

    def handle_break(self, ctx):
        # disable all breakpoints so that we don't get stuck
        for i in range(5):
            self.u.msr(DBGBCRn_EL1(i), 0)

        # we'll need to single step to enable these breakpoints again
        self.u.msr(MDSCR_EL1, MDSCR(SS=1, MDE=1).value)
        self.ctx.spsr.SS = 1

        if ctx.elr in self._bp_hooks:
            if self._bp_hooks[ctx.elr](ctx):
                return True

    def handle_watch(self, ctx):
        # disable all watchpoints so that we don't get stuck
        for i in range(len(self._wps)):
            self.u.msr(DBGWCRn_EL1(i), 0)

        # we'll need to single step to enable these watchpoints again
        self.u.msr(MDSCR_EL1, MDSCR(SS=1, MDE=1).value)
        self.ctx.spsr.SS = 1

    def add_hvcall(self, callid, handler):
        self.hvcall_handlers[callid] = handler

    def handle_brk(self, ctx):
        iss = ctx.esr.ISS
        if iss != 0x4242:
            return self._lower()

        # HV call from EL0/1
        callid = ctx.regs[0]
        handler = self.hvcall_handlers.get(callid, None)
        if handler is None:
            self.log(f"Undefined HV call #{callid}")
            return False

        ok = handler(ctx)
        if ok:
            ctx.elr += 4
        return ok

    def handle_dabort(self, ctx):
        insn = self.p.read32(ctx.elr_phys)
        far_phys = self.p.hv_translate(ctx.far, True, False)

        if insn & 0x3b200c00 == 0x38200000:
            page = far_phys & ~0x3fff

            before = self.p.read32(far_phys)
            self.map_hw(page, page, 0x4000)
            r0b = self.ctx.regs[0]
            self.log(f"-ELR={self.ctx.elr:#x} LR={self.ctx.regs[30]:#x}")
            self.step()
            self.log(f"+ELR={self.ctx.elr:#x}")
            r0a = self.ctx.regs[0]
            self.dirty_maps.set(irange(page, 0x4000))
            self.pt_update()
            after = self.p.read32(far_phys)
            self.log(f"Unhandled atomic: @{far_phys:#x} {before:#x} -> {after:#x} | r0={r0b:#x} -> {r0a:#x}")
            return True

        if insn & 0x3f000000 == 0x08000000:
            page = far_phys & ~0x3fff
            before = self.p.read32(far_phys)
            self.map_hw(page, page, 0x4000)
            r0b = self.ctx.regs[0]
            self.log(f"-ELR={self.ctx.elr:#x} LR={self.ctx.regs[30]:#x}")
            self.step()
            self.log(f"+ELR={self.ctx.elr:#x}")
            r0a = self.ctx.regs[0]
            self.dirty_maps.set(irange(page, 0x4000))
            self.pt_update()
            after = self.p.read32(far_phys)
            self.log(f"Unhandled exclusive: @{far_phys:#x} {before:#x} -> {after:#x} | r0={r0b:#x} -> {r0a:#x}")
            return True

    def handle_sync(self, ctx):
        if ctx.esr.EC == ESR_EC.MSR:
            return self.handle_msr(ctx)

        if ctx.esr.EC == ESR_EC.IMPDEF:
            return self.handle_impdef(ctx)

        if ctx.esr.EC == ESR_EC.HVC:
            return self.handle_hvc(ctx)

        if ctx.esr.EC == ESR_EC.SSTEP_LOWER:
            return self.handle_step(ctx)

        if ctx.esr.EC == ESR_EC.BKPT_LOWER:
            return self.handle_break(ctx)

        if ctx.esr.EC == ESR_EC.WATCH_LOWER:
            return self.handle_watch(ctx)

        if ctx.esr.EC == ESR_EC.BRK:
            return self.handle_brk(ctx)

        if ctx.esr.EC == ESR_EC.DABORT_LOWER:
            return self.handle_dabort(ctx)

    def _load_context(self):
        self._info_data = self.iface.readmem(self.exc_info, ExcInfo.sizeof())
        self.ctx = ExcInfo.parse(self._info_data)
        return self.ctx

    def _commit_context(self):
        new_info = ExcInfo.build(self.ctx)
        if new_info != self._info_data:
            self.iface.writemem(self.exc_info, new_info)
            self._info_data = new_info

    def handle_exception(self, reason, code, info):
        self.exc_info = info
        self.exc_reason = reason
        if reason in (START.EXCEPTION_LOWER, START.EXCEPTION):
            code = EXC(code)
        elif reason == START.HV:
            code = HV_EVENT(code)
        self.exc_code = code
        self.is_fault = reason == START.EXCEPTION_LOWER and code in (EXC.SYNC, EXC.SERROR)

        # Nested context switch is handled by the caller
        if self.switching_context:
            self.switching_context = False
            return

        self._in_handler = True

        ctx = self._load_context()
        self.exc_orig_cpu = self.ctx.cpu_id

        handled = False
        user_interrupt = False

        try:
            if reason == START.EXCEPTION_LOWER:
                if code == EXC.SYNC:
                    handled = self.handle_sync(ctx)
                elif code == EXC.FIQ:
                    self.u.msr(CNTV_CTL_EL0, 0)
                    self.u.print_context(ctx, False, sym=self.get_sym)
                    handled = True
            elif reason == START.HV:
                code = HV_EVENT(code)
                if code == HV_EVENT.HOOK_VM:
                    handled = self.handle_vm_hook(ctx)
                elif code == HV_EVENT.USER_INTERRUPT:
                    handled = True
                    user_interrupt = True
        except Exception as e:
            self.log(f"Python exception while handling guest exception:")
            traceback.print_exc()

        if handled:
            ret = EXC_RET.HANDLED
            if self._sigint_pending:
                self.update_pac_mask()
                self.log("User interrupt")
        else:
            self.log(f"Guest exception: {reason.name}/{code.name}")
            self.update_pac_mask()
            self.u.print_context(ctx, self.is_fault, sym=self.get_sym)

        if self._sigint_pending or not handled or user_interrupt:
            self._sigint_pending = False

            signal.signal(signal.SIGINT, self.default_sigint)
            ret = self.run_shell("Entering hypervisor shell", "Returning from exception")
            signal.signal(signal.SIGINT, self._handle_sigint)

            if ret is None:
                ret = EXC_RET.HANDLED

        self.pt_update()

        self._commit_context()
        self.ctx = None
        self.exc_orig_cpu = None
        self.p.exit(ret)

        self._in_handler = False
        if self._sigint_pending:
            self._handle_sigint()

    def handle_bark(self, reason, code, info):
        self._in_handler = True
        self._sigint_pending = False

        signal.signal(signal.SIGINT, self.default_sigint)
        ret = self.run_shell("Entering panic shell", "Exiting")
        signal.signal(signal.SIGINT, self._handle_sigint)

        self.p.exit(0)

    def attach_virtio(self, dev, base=None, irq=None, verbose=False):
        if base is None:
            base = alloc_mmio_base(self.adt, 0x1000)
        if irq is None:
            irq = alloc_aic_irq(self.adt)

        data = dev.config_data
        data_base = self.u.heap.malloc(len(data))
        self.iface.writemem(data_base, data)

        config = VirtioConfig.build({
            "irq": irq,
            "devid": dev.devid,
            "feats": dev.feats,
            "num_qus": dev.num_qus,
            "data": data_base,
            "data_len": len(data),
            "verbose": verbose,
        })

        config_base = self.u.heap.malloc(len(config))
        self.iface.writemem(config_base, config)

        name = None
        for i in range(16):
            n = "/arm-io/virtio%d" % i
            if n not in self.adt:
                name = n
                break
        if name is None:
            raise ValueError("Too many virtios in ADT")

        print(f"Adding {n} @ 0x{base:x}, irq {irq}")

        node = self.adt.create_node(name)
        node.reg = [Container(addr=node.to_bus_addr(base), size=0x1000)]
        node.interrupt_parent = getattr(self.adt["/arm-io/aic"], "AAPL,phandle")
        node.interrupts = (irq,)
        node.compatible = ["virtio,mmio"]

        self.p.hv_map_virtio(base, config_base)
        self.add_tracer(irange(base, 0x1000), "VIRTIO", TraceMode.RESERVED)

        dev.base = base
        dev.hv = self
        self.virtio_devs[base] = dev

    def handle_virtio(self, reason, code, info):
        ctx = self.iface.readstruct(info, ExcInfo)
        self.virtio_ctx = info = self.iface.readstruct(ctx.data, VirtioExcInfo)

        try:
            handled = self.virtio_devs[info.devbase].handle_exc(info)
        except:
            self.log(f"Python exception from within virtio handler")
            traceback.print_exc()
            handled = False

        if not handled:
            signal.signal(signal.SIGINT, self.default_sigint)
            self.run_shell("Entering hypervisor shell", "Returning")
            signal.signal(signal.SIGINT, self._handle_sigint)

        self.p.exit(EXC_RET.HANDLED)

    def skip(self):
        self.ctx.elr += 4
        self.cont()

    def cont(self):
        os.kill(os.getpid(), signal.SIGUSR1)

    def _lower(self):
        if not self.is_fault:
            print("Cannot lower non-fault exception")
            return False

        self.u.msr(ELR_EL12, self.ctx.elr)
        self.u.msr(SPSR_EL12, self.ctx.spsr.value)
        self.u.msr(ESR_EL12, self.ctx.esr.value)
        self.u.msr(FAR_EL12, self.ctx.far)

        exc_off = 0x80 * self.exc_code

        if self.ctx.spsr.M == SPSR_M.EL0t:
            exc_off += 0x400
        elif self.ctx.spsr.M == SPSR_M.EL1t:
            pass
        elif self.ctx.spsr.M == SPSR_M.EL1h:
            exc_off += 0x200
        else:
            print(f"Unknown exception level {self.ctx.spsr.M}")
            return False

        self.ctx.spsr.M = SPSR_M.EL1h
        self.ctx.spsr.D = 1
        self.ctx.spsr.A = 1
        self.ctx.spsr.I = 1
        self.ctx.spsr.F = 1
        self.ctx.elr = self.u.mrs(VBAR_EL12) + exc_off

        return True

    def lower(self, step=False):
        self.cpu() # Return to exception CPU

        if not self._lower():
            return
        elif step:
            self.step()
        else:
            self.cont()

    def step(self):
        self.u.msr(MDSCR_EL1, MDSCR(SS=1, MDE=1).value)
        self.ctx.spsr.SS = 1
        self.p.hv_pin_cpu(self.ctx.cpu_id)
        self._switch_context()
        self.p.hv_pin_cpu(0xffffffffffffffff)

    def _switch_context(self, exit=EXC_RET.HANDLED):
        # Flush current CPU context out to HV
        self._commit_context()
        self.exc_info = None
        self.ctx = None

        self.switching_context = True
        # Exit out of the proxy
        self.p.exit(exit)
        # Wait for next proxy entry
        self.iface.wait_and_handle_boot()
        if self.switching_context:
            raise Exception(f"Failed to switch context")

        # Fetch new context
        self._load_context()

    def cpu(self, cpu=None):
        if cpu is None:
            cpu = self.exc_orig_cpu
        if cpu == self.ctx.cpu_id:
            return

        if not self.p.hv_switch_cpu(cpu):
            raise ValueError(f"Invalid or inactive CPU #{cpu}")

        self._switch_context()
        if self.ctx.cpu_id != cpu:
            raise Exception(f"Switching to CPU #{cpu} but ended on #{self.ctx.cpu_id}")

    def add_hw_bp(self, vaddr, hook=None):
        if None not in self._bps:
            raise ValueError("Cannot add more HW breakpoints")

        i = self._bps.index(None)
        cpu_id = self.ctx.cpu_id
        try:
            for cpu in self.cpus():
                self.u.msr(DBGBCRn_EL1(i), DBGBCR(E=1, PMC=0b11, BAS=0xf).value)
                self.u.msr(DBGBVRn_EL1(i), vaddr)
        finally:
            self.cpu(cpu_id)
        self._bps[i] = vaddr
        if hook is not None:
            self._bp_hooks[vaddr] = hook

    def remove_hw_bp(self, vaddr):
        idx = self._bps.index(vaddr)
        self._bps[idx] = None
        cpu_id = self.ctx.cpu_id
        try:
            for cpu in self.cpus():
                self.u.msr(DBGBCRn_EL1(idx), 0)
                self.u.msr(DBGBVRn_EL1(idx), 0)
        finally:
            self.cpu(cpu_id)
        if vaddr in self._bp_hooks:
            del self._bp_hooks[vaddr]

    def add_sym_bp(self, name, hook=None):
        return self.add_hw_bp(self.resolve_symbol(name), hook=hook)

    def remove_sym_bp(self, name):
        return self.remove_hw_bp(self.resolve_symbol(name))

    def clear_hw_bps(self):
        for vaddr in self._bps:
            self.remove_hw_bp(vaddr)

    def add_hw_wp(self, vaddr, bas, lsc):
        for i, i_vaddr in enumerate(self._wps):
            if i_vaddr is None:
                self._wps[i] = vaddr
                self._wpcs[i] = DBGWCR(E=1, PAC=0b11, BAS=bas, LSC=lsc).value
                cpu_id = self.ctx.cpu_id
                try:
                    for cpu in self.cpus():
                        self.u.msr(DBGWCRn_EL1(i), self._wpcs[i])
                        self.u.msr(DBGWVRn_EL1(i), vaddr)
                finally:
                    self.cpu(cpu_id)
                return
        raise ValueError("Cannot add more HW watchpoints")

    def get_wp_bas(self, vaddr):
        for i, i_vaddr in enumerate(self._wps):
            if i_vaddr == vaddr:
                return self._wpcs[i].BAS

    def remove_hw_wp(self, vaddr):
        idx = self._wps.index(vaddr)
        self._wps[idx] = None
        self._wpcs[idx] = 0
        cpu_id = self.ctx.cpu_id
        try:
            for cpu in self.cpus():
                self.u.msr(DBGWCRn_EL1(idx), 0)
                self.u.msr(DBGWVRn_EL1(idx), 0)
        finally:
            self.cpu(cpu_id)

    def exit(self):
        os.kill(os.getpid(), signal.SIGUSR2)

    def reboot(self):
        print("Hard rebooting the system")
        self.p.reboot()
        sys.exit(0)

    def hvc(self, arg):
        assert 0 <= arg <= 0xffff
        return 0xd4000002 | (arg << 5)

    def decode_dbg_panic(self):
        xnutools.decode_debugger_state(self.u, self.ctx)

    def decode_panic_call(self):
        xnutools.decode_panic_call(self.u, self.ctx)

    def context(self):
        f = f" (orig: #{self.exc_orig_cpu})" if self.ctx.cpu_id != self.exc_orig_cpu else ""
        print(f"  == On CPU #{self.ctx.cpu_id}{f} ==")
        print(f"  Reason: {self.exc_reason.name}/{self.exc_code.name}")
        self.u.print_context(self.ctx, self.is_fault, sym=self.get_sym)

    def bt(self, frame=None, lr=None):
        if frame is None:
            frame = self.ctx.regs[29]
        if lr is None:
            lr = self.unpac(self.ctx.elr) + 4

        print("Stack trace:")
        frames = set()
        while frame:
            if frame in frames:
                print("Stack loop detected!")
                break
            frames.add(frame)
            print(f" - {self.addr(lr - 4)}")
            lrp = self.p.hv_translate(frame + 8)
            fpp = self.p.hv_translate(frame)
            if not fpp:
                break
            lr = self.unpac(self.p.read64(lrp))
            frame = self.p.read64(fpp)

    def cpus(self):
        for i in sorted(self.started_cpus):
            self.cpu(i)
            yield i

    def patch_exception_handling(self):
        if self.ctx.cpu_id != 0:
            return

        if self.want_vbar is not None:
            vbar = self.want_vbar
        else:
            vbar = self.u.mrs(VBAR_EL12)

        if vbar == self.vbar_el1:
            return

        if vbar == 0:
            return

        if self.u.mrs(SCTLR_EL12) & 1:
            vbar_phys = self.p.hv_translate(vbar, False, False)
            if vbar_phys == 0:
                self.log(f"VBAR vaddr 0x{vbar:x} translation failed!")
                if self.vbar_el1 is not None:
                    self.want_vbar = vbar
                    self.u.msr(VBAR_EL12, self.vbar_el1)
                return
        else:
            if vbar & (1 << 63):
                self.log(f"VBAR vaddr 0x{vbar:x} without translation enabled")
                if self.vbar_el1 is not None:
                    self.want_vbar = vbar
                    self.u.msr(VBAR_EL12, self.vbar_el1)
                return

            vbar_phys = vbar

        if self.want_vbar is not None:
            self.want_vbar = None
            self.u.msr(VBAR_EL12, vbar)

        self.log(f"New VBAR paddr: 0x{vbar_phys:x}")

        #for i in range(16):
        for i in [0, 3, 4, 7, 8, 11, 12, 15]:
            idx = 0
            addr = vbar_phys + 0x80 * i
            orig = self.p.read32(addr)
            if (orig & 0xfc000000) != 0x14000000:
                self.log(f"Unknown vector #{i}:\n")
                self.u.disassemble_at(addr, 16)
            else:
                idx = len(self.vectors)
                delta = orig & 0x3ffffff
                if delta == 0:
                    target = None
                    self.log(f"Vector #{i}: Loop\n")
                else:
                    target = (delta << 2) + vbar + 0x80 * i
                    self.log(f"Vector #{i}: 0x{target:x}\n")
                self.vectors.append((i, target))
                self.u.disassemble_at(addr, 16)
            self.p.write32(addr, self.hvc(idx))

        self.p.dc_cvau(vbar_phys, 0x800)
        self.p.ic_ivau(vbar_phys, 0x800)

        self.vbar_el1 = vbar

    def set_logfile(self, fd):
        self.print_tracer.log_file = fd

    def init(self):
        self.adt = load_adt(self.u.get_adt())
        self.iodev = self.p.iodev_whoami()
        self.tba = self.u.ba.copy()
        self.device_addr_tbl = self.adt.build_addr_lookup()
        self.print_tracer = trace.PrintTracer(self, self.device_addr_tbl)

        # disable unused USB iodev early so interrupts can be reenabled in hv_init()
        for iodev in IODEV:
            if iodev >= IODEV.USB0 and iodev != self.iodev:
                print(f"Disable iodev {iodev!s}")
                self.p.iodev_set_usage(iodev, 0)

        print("Initializing hypervisor over iodev %s" % self.iodev)
        self.p.hv_init()

        self.iface.set_handler(START.EXCEPTION_LOWER, EXC.SYNC, self.handle_exception)
        self.iface.set_handler(START.EXCEPTION_LOWER, EXC.IRQ, self.handle_exception)
        self.iface.set_handler(START.EXCEPTION_LOWER, EXC.FIQ, self.handle_exception)
        self.iface.set_handler(START.EXCEPTION_LOWER, EXC.SERROR, self.handle_exception)
        self.iface.set_handler(START.EXCEPTION, EXC.FIQ, self.handle_exception)
        self.iface.set_handler(START.HV, HV_EVENT.USER_INTERRUPT, self.handle_exception)
        self.iface.set_handler(START.HV, HV_EVENT.HOOK_VM, self.handle_exception)
        self.iface.set_handler(START.HV, HV_EVENT.VTIMER, self.handle_exception)
        self.iface.set_handler(START.HV, HV_EVENT.WDT_BARK, self.handle_bark)
        self.iface.set_handler(START.HV, HV_EVENT.CPU_SWITCH, self.handle_exception)
        self.iface.set_handler(START.HV, HV_EVENT.VIRTIO, self.handle_virtio)
        self.iface.set_handler(START.HV, HV_EVENT.PANIC, self.handle_bark)
        self.iface.set_event_handler(EVENT.MMIOTRACE, self.handle_mmiotrace)
        self.iface.set_event_handler(EVENT.IRQTRACE, self.handle_irqtrace)

        # Map MMIO ranges as HW by default
        for r in self.adt["/arm-io"].ranges:
            print(f"Mapping MMIO range: {r.parent_addr:#x} .. {r.parent_addr + r.size:#x}")
            self.add_tracer(irange(r.parent_addr, r.size), "HW", TraceMode.OFF)

        hcr = HCR(self.u.mrs(HCR_EL2))
        if self.novm:
            hcr.VM = 0
            hcr.AMO = 0
        else:
            hcr.TACR = 1
        hcr.TIDCP = 0
        hcr.TVM = 0
        hcr.FMO = 1
        hcr.IMO = 0
        hcr.TTLBOS = 1
        self.u.msr(HCR_EL2, hcr.value)

        # Trap dangerous things
        hacr = HACR(0)
        if not self.novm:
            #hacr.TRAP_CPU_EXT = 1
            #hacr.TRAP_SPRR = 1
            #hacr.TRAP_GXF = 1
            hacr.TRAP_CTRR = 1
            hacr.TRAP_EHID = 1
            hacr.TRAP_HID = 1
            hacr.TRAP_ACC = 1
            hacr.TRAP_IPI = 1
            hacr.TRAP_SERROR_INFO = 1 # M1RACLES mitigation
            hacr.TRAP_PM = 1
        self.u.msr(HACR_EL2, hacr.value)

        # enable and route debug exceptions to EL2
        mdcr = MDCR(0)
        mdcr.TDE = 1
        mdcr.TDA = 1
        mdcr.TDOSA = 1
        mdcr.TDRA = 1
        self.u.msr(MDCR_EL2, mdcr.value)
        self.u.msr(MDSCR_EL1, MDSCR(MDE=1).value)

        # Enable AMX
        amx_ctl = AMX_CONFIG(self.u.mrs(AMX_CONFIG_EL1))
        amx_ctl.EN_EL1 = 1
        self.u.msr(AMX_CONFIG_EL1, amx_ctl.value)

        # Set guest AP keys
        self.u.msr(VMKEYLO_EL2, 0x4E7672476F6E6147)
        self.u.msr(VMKEYHI_EL2, 0x697665596F755570)
        self.u.msr(APSTS_EL12, 1)

        self.map_vuart()

        # ACTLR depends on the CPU part
        part = MIDR(self.u.mrs(MIDR_EL1)).PART
        actlr_el12 = ACTLR_EL12 if part >= MIDR_PART.T8110_BLIZZARD else ACTLR_EL12_PRE

        actlr = ACTLR(self.u.mrs(actlr_el12))
        actlr.EnMDSB = 1
        self.u.msr(actlr_el12, actlr.value)

        self.setup_adt()

    def map_vuart(self):
        node = base = self.adt["/arm-io/uart0"]
        base = node.get_reg(0)[0]

        zone = irange(base, 0x4000)
        irq = node.interrupts[0]
        self.p.hv_map_vuart(base, irq, self.iodev)
        self.add_tracer(zone, "VUART", TraceMode.RESERVED)

    def map_essential(self):
        # Things we always map/take over, for the hypervisor to work
        _pmgr = {}

        def wh(base, off, data, width):
            self.log(f"PMGR W {base:x}+{off:x}:{width} = 0x{data:x}: Dangerous write")
            self.p.mask32(base + off, 0x3ff, (data | 0xf) & ~(0x80000400))
            _pmgr[base + off] = (data & 0xfffffc0f) | ((data & 0xf) << 4)

        def rh(base, off, width):
            data = self.p.read32(base + off)
            ret = _pmgr.setdefault(base + off, data)
            self.log(f"PMGR R {base:x}+{off:x}:{width} = 0x{data:x} -> 0x{ret:x}")
            return ret

        atc = f"ATC{self.iodev - IODEV.USB0}_USB"
        atc_aon = f"ATC{self.iodev - IODEV.USB0}_USB_AON"

        hook_devs = ["UART0", atc, atc_aon]

        pmgr = self.adt["/arm-io/pmgr"]
        dev_by_name = {dev.name: dev for dev in pmgr.devices}
        dev_by_id = {dev.id: dev for dev in pmgr.devices}

        pmgr_hooks = []

        def hook_pmgr_dev(dev):
            ps = pmgr.ps_regs[dev.psreg]
            if dev.psidx or dev.psreg:
                addr = pmgr.get_reg(ps.reg)[0] + ps.offset + dev.psidx * 8
                pmgr_hooks.append(addr)
                for idx in dev.parents:
                    if idx in dev_by_id:
                        hook_pmgr_dev(dev_by_id[idx])

        for name in hook_devs:
            dev = dev_by_name[name]
            hook_pmgr_dev(dev)

        pmgr0_start = pmgr.get_reg(0)[0]

        for addr in pmgr_hooks:
            self.map_hook(addr, 4, write=wh, read=rh)
            #TODO : turn into a real tracer
            self.add_tracer(irange(addr, 4), "PMGR HACK", TraceMode.RESERVED)

        pg_overrides = {
            0x23d29c05c: 0xc000000,
            0x23d29c044: 0xc000000,
        }

        for addr in pg_overrides:
            self.map_hook(addr, 4, read=lambda base, off, width: pg_overrides[base + off])
            self.add_tracer(irange(addr, 4), "PMGR HACK", TraceMode.RESERVED)

        cpu_hack = [
            # 0x210e20020,
            # 0x211e20020,
            # 0x212e20020,
        ]

        def wh(base, off, data, width):
            if isinstance(data, list):
                data = data[0]
            self.log(f"CPU W {base:x}+{off:x}:{width} = 0x{data:x}: Dangerous write")

        for addr in cpu_hack:
            self.map_hook(addr, 8, write=wh)
            self.add_tracer(irange(addr, 8), "CPU HACK", TraceMode.RESERVED)

        def cpu_state_rh(base, off, width):
            data = ret = self.p.read64(base + off)
            die = base // 0x20_0000_0000
            cluster = (base >> 24) & 0xf
            cpu = (base >> 20) & 0xf
            for i, j in self.started_cpus.items():
                if j == (die, cluster, cpu):
                    break
            else:
                ret &= ~0xff
            self.log(f"CPU STATE R {base:x}+{off:x}:{width} = 0x{data:x} -> 0x{ret:x}")
            return ret

        def cpustart_wh(base, off, data, width):
            self.log(f"CPUSTART W {base:x}+{off:x}:{width} = 0x{data:x}")
            if off >= 8:
                assert width == 32
                die = base // 0x20_0000_0000
                cluster = (off - 8) // 4
                for i in range(32):
                    if data & (1 << i):
                        self.start_secondary(die, cluster, i)
                        cpu_state = 0x210050100 | (die << 27) | (cluster << 24) | (i << 20)
                        self.map_hook(cpu_state, 8, read=cpu_state_rh)
                        self.add_tracer(irange(addr, 8), "CPU STATE HACK", TraceMode.RESERVED)

        die_count = self.adt["/arm-io"].die_count if hasattr(self.adt["/arm-io"], "die-count") else 1

        for die in range(0, die_count):
            chip_id = self.u.adt["/chosen"].chip_id
            if chip_id in (0x8103, 0x6000, 0x6001, 0x6002):
                cpu_start = 0x54000 + die * 0x20_0000_0000
            elif chip_id in (0x8112, 0x8122):
                cpu_start = 0x34000 + die * 0x20_0000_0000
            elif chip_id in (0x6020, 0x6021, 0x6022):
                cpu_start = 0x28000 + die * 0x20_0000_0000
            elif chip_id in (0x6031,):
                cpu_start = 0x88000 + die * 0x20_0000_0000
            else:
                self.log("CPUSTART unknown for this SoC!")
                break

            zone = irange(pmgr0_start + cpu_start, 0x20)
            self.map_hook(pmgr0_start + cpu_start, 0x20, write=cpustart_wh)
            self.add_tracer(zone, "CPU_START", TraceMode.RESERVED)

    def start_secondary(self, die, cluster, cpu):
        self.log(f"Starting guest secondary {die}:{cluster}:{cpu}")

        for node in list(self.adt["cpus"]):
            if ((die << 11) | (cluster << 8) | cpu) == node.reg:
                break
        else:
            self.log("CPU not found!")
            return

        entry = self.p.read64(node.cpu_impl_reg[0]) & 0xfffffffffff
        index = node.cpu_id
        self.log(f" CPU #{index}: RVBAR = {entry:#x}")

        self.sysreg[index] = {}
        self.started_cpus[index] = (die, cluster, cpu)
        self.p.hv_start_secondary(index, entry)

    def setup_adt(self):
        self.adt["product"].product_name += " on m1n1 hypervisor"
        self.adt["product"].product_description += " on m1n1 hypervisor"
        soc_name = "Virtual " + self.adt["product"].product_soc_name + " on m1n1 hypervisor"
        self.adt["product"].product_soc_name = soc_name

        if self.iodev >= IODEV.USB0:
            idx = self.iodev - IODEV.USB0
            for prefix in ("/arm-io/dart-usb%d",
                           "/arm-io/atc-phy%d",
                           "/arm-io/usb-drd%d",
                           "/arm-io/acio%d",
                           "/arm-io/acio-cpu%d",
                           "/arm-io/dart-acio%d",
                           "/arm-io/apciec%d",
                           "/arm-io/dart-apciec%d",
                           "/arm-io/apciec%d-piodma",
                           "/arm-io/i2c0/hpmBusManager/hpm%d",
                           "/arm-io/atc%d-dpxbar",
                           "/arm-io/atc%d-dpphy",
                           "/arm-io/atc%d-dpin0",
                           "/arm-io/atc%d-dpin1",
                           "/arm-io/atc-phy%d",
                          ):
                name = prefix % idx
                print(f"Removing ADT node {name}")
                try:
                    del self.adt[name]
                except KeyError:
                    pass

        if self.wdt_cpu is not None:
            name = f"/cpus/cpu{self.wdt_cpu}"
            print(f"Removing ADT node {name}")
            try:
                del self.adt[name]
            except KeyError:
                pass

        if not self.smp:
            for cpu in list(self.adt["cpus"]):
                if cpu.state != "running":
                    print(f"Removing ADT node {cpu._path}")
                    try:
                        del self.adt["cpus"][cpu.name]
                    except KeyError:
                        pass

    def set_bootargs(self, boot_args):
        if "-v" in boot_args.split():
            self.tba.video.display = 0
        else:
            self.tba.video.display = 1
        print(f"Setting boot arguments to {boot_args!r}")
        self.tba.cmdline = boot_args

    def unmap_carveouts(self):
        print(f"Unmapping TZ carveouts...")
        carveout_p = self.p.mcc_get_carveouts()
        while True:
            base = self.p.read64(carveout_p)
            size = self.p.read64(carveout_p + 8)
            if not base:
                break
            print(f"  Unmap [{base:#x}..{base + size - 1:#x}]")
            self.del_tracer(irange(base, size), "RAM-LOW")
            self.del_tracer(irange(base, size), "RAM-HIGH")
            carveout_p += 16

    def enable_time_stealing(self):
        self.p.hv_set_time_stealing(True)

    def disable_time_stealing(self):
        self.p.hv_set_time_stealing(False)


    def load_raw(self, image, entryoffset=0x800, use_xnu_symbols=False, vmin=0):
        sepfw_start, sepfw_length = self.u.adt["chosen"]["memory-map"].SEPFW
        tc_start, tc_size = self.u.adt["chosen"]["memory-map"].TrustCache
        if hasattr(self.u.adt["chosen"]["memory-map"], "preoslog"):
            preoslog_start, preoslog_size = self.u.adt["chosen"]["memory-map"].preoslog
        else:
            preoslog_size = 0

        image_size = align(len(image))
        sepfw_off = image_size
        image_size += align(sepfw_length)
        preoslog_off = image_size
        image_size += preoslog_size
        self.bootargs_off = image_size
        bootargs_size = 0x4000
        image_size += bootargs_size

        print(f"Total region size: 0x{image_size:x} bytes")

        self.phys_base = phys_base = guest_base = self.u.heap_top
        self.ram_base = self.phys_base & ~0xffffffff
        self.ram_size = self.u.ba.mem_size_actual
        guest_base += 16 << 20 # ensure guest starts within a 16MB aligned region of mapped RAM
        self.adt_base = guest_base
        guest_base += align(self.u.ba.devtree_size)
        tc_base = guest_base
        guest_base += align(tc_size)
        self.guest_base = guest_base
        mem_top = self.u.ba.phys_base + self.u.ba.mem_size
        mem_size = mem_top - phys_base

        print(f"Physical memory: 0x{phys_base:x} .. 0x{mem_top:x}")
        print(f"Guest region start: 0x{guest_base:x}")
        
        self.entry = guest_base + entryoffset

        print(f"Mapping guest physical memory...")
        self.add_tracer(irange(self.ram_base, self.u.ba.phys_base - self.ram_base), "RAM-LOW", TraceMode.OFF)
        self.add_tracer(irange(phys_base, self.u.ba.mem_size_actual - phys_base + self.ram_base), "RAM-HIGH", TraceMode.OFF)
        self.unmap_carveouts()

        print(f"Loading kernel image (0x{len(image):x} bytes)...")
        self.u.compressed_writemem(guest_base, image, True)
        self.p.dc_cvau(guest_base, len(image))
        self.p.ic_ivau(guest_base, len(image))

        print(f"Copying SEPFW (0x{sepfw_length:x} bytes)...")
        self.p.memcpy8(guest_base + sepfw_off, sepfw_start, sepfw_length)

        print(f"Copying TrustCache (0x{tc_size:x} bytes)...")
        self.p.memcpy8(tc_base, tc_start, tc_size)

        if hasattr(self.u.adt["chosen"]["memory-map"], "preoslog"):
            print(f"Copying preoslog (0x{preoslog_size:x} bytes)...")
            self.p.memcpy8(guest_base + preoslog_off, preoslog_start, preoslog_size)

        print(f"Adjusting addresses in ADT...")
        self.adt["chosen"]["memory-map"].SEPFW = (guest_base + sepfw_off, sepfw_length)
        self.adt["chosen"]["memory-map"].TrustCache = (tc_base, tc_size)
        self.adt["chosen"]["memory-map"].DeviceTree = (self.adt_base, align(self.u.ba.devtree_size))
        self.adt["chosen"]["memory-map"].BootArgs = (guest_base + self.bootargs_off, bootargs_size)
        if hasattr(self.u.adt["chosen"]["memory-map"], "preoslog"):
            self.adt["chosen"]["memory-map"].preoslog = (guest_base + preoslog_off, preoslog_size)
        if hasattr(self.u.adt["chosen"]["memory-map"], "Kernel_mach__header"):
            self.adt["chosen"]["memory-map"].Kernel_mach__header = (guest_base, 0)

        def remove_oslog(node):
            names = node.segment_names.split(";")
            try:
                idx = names.index("__OS_LOG")
            except ValueError:
                return
            print(f"Removing __OS_LOG from {node.name}")
            names = names[:idx] + names[idx + 1:]
            node.segment_names = ";".join(names)
            node.segment_ranges = node.segment_ranges[:idx * 32] + node.segment_ranges[32 + idx * 32: ]

        for node in self.adt["/arm-io"]:
            if hasattr(node, "segment_names"):
                remove_oslog(node)
            for nub in node:
                if hasattr(nub, "segment_names"):
                    remove_oslog(nub)

        print(f"Setting up bootargs at 0x{guest_base + self.bootargs_off:x}...")

        self.tba.mem_size = mem_size
        self.tba.phys_base = phys_base
        self.tba.virt_base = 0xfffffe0010000000 + (phys_base & (32 * 1024 * 1024 - 1))
        self.tba.devtree = self.adt_base - phys_base + self.tba.virt_base
        self.tba.top_of_kernel_data = guest_base + image_size

        if use_xnu_symbols == True:
            self.sym_offset = vmin - guest_base + self.tba.phys_base - self.tba.virt_base

        if self.tba.revision <= 1:
            self.iface.writemem(guest_base + self.bootargs_off, BootArgs_r1.build(self.tba))
        elif self.tba.revision == 2:
            self.iface.writemem(guest_base + self.bootargs_off, BootArgs_r2.build(self.tba))
        elif self.tba.revision == 3:
            self.iface.writemem(guest_base + self.bootargs_off, BootArgs_r3.build(self.tba))

        print("Setting secondary CPU RVBARs...")
        rvbar = self.entry & ~0xfff
        for cpu in self.adt["cpus"]:
            if cpu.state == "running":
                continue
            addr, size = cpu.cpu_impl_reg
            print(f"  {cpu.name}: [0x{addr:x}] = 0x{rvbar:x}")
            self.p.write64(addr, rvbar)

    def _load_macho_symbols(self):
        self.symbol_dict = self.macho.symbols
        self.symbols = [(v, k) for k, v in self.macho.symbols.items()]
        self.symbols.sort()

    def load_macho(self, data, symfile=None):
        if isinstance(data, str):
            data = open(data, "rb")

        self.xnu_mode = True

        self.macho = macho = MachO(data)
        if symfile is not None:
            if isinstance(symfile, str):
                symfile = open(symfile, "rb")
            syms = MachO(symfile)
            macho.add_symbols("com.apple.kernel", syms)

        self._load_macho_symbols()

        def load_hook(data, segname, size, fileoff, dest):
            if segname != "__TEXT_EXEC":
                return data

            print(f"Patching segment {segname}...")

            a = array.array("I", data)

            output = []

            p = 0
            while (p := data.find(b"\x20\x00", p)) != -1:
                if (p & 3) != 2:
                    p += 1
                    continue

                opcode = a[p // 4]
                inst = self.hvc((opcode & 0xffff))
                off = fileoff + (p & ~3)
                if off >= 0xbfcfc0:
                    print(f"  0x{off:x}: 0x{opcode:04x} -> hvc 0x{opcode:x} (0x{inst:x})")
                    a[p // 4] = inst
                p += 4

            print("Done.")
            return a.tobytes()

        #image = macho.prepare_image(load_hook)
        image = macho.prepare_image()
        self.load_raw(image, entryoffset=(macho.entry - macho.vmin), use_xnu_symbols=self.xnu_mode and symfile is not None, vmin=macho.vmin)

    def update_pac_mask(self):
        tcr = TCR(self.u.mrs(TCR_EL12))
        valid_bits = (1 << (64 - tcr.T1SZ)) - 1
        self.pac_mask = 0xffffffffffffffff & ~valid_bits
        valid_bits = (1 << (64 - tcr.T0SZ)) - 1
        self.user_pac_mask = 0xffffffffffffffff & ~valid_bits

    def unpac(self, v):
        if v & (1 << 55):
            return v | self.pac_mask
        else:
            return v & ~self.user_pac_mask

    def load_system_map(self, path):
        # Assume Linux
        self.sym_offset = 0
        self.xnu_mode = False
        self.symbols = []
        self.symbol_dict = {}
        with open(path) as fd:
            for line in fd.readlines():
                addr, t, name = line.split()
                addr = int(addr, 16)
                self.symbols.append((addr, name))
                self.symbol_dict[name] = addr
        self.symbols.sort()

    def add_kext_symbols(self, kext, demangle=False):
        info_plist = plistlib.load(open(f"{kext}/Contents/Info.plist", "rb"))
        identifier = info_plist["CFBundleIdentifier"]
        name = info_plist["CFBundleName"]
        macho = MachO(open(f"{kext}/Contents/MacOS/{name}", "rb"))
        self.macho.add_symbols(identifier, macho, demangle=demangle)
        self._load_macho_symbols()

    def _handle_sigint(self, signal=None, stack=None):
        self._sigint_pending = True
        self.interrupt()

    def interrupt(self):
        if self._in_handler:
            return

        # Kick the proxy to break out of the hypervisor
        self.iface.dev.write(b"!")

    def run_script(self, path):
        new_locals = runpy.run_path(path, init_globals=self.shell_locals, run_name="<hv_script>")
        self.shell_locals.clear()
        self.shell_locals.update(new_locals)

    def run_code(self, code):
        exec(code, self.shell_locals)

    def start(self):
        print("Disabling other iodevs...")
        for iodev in IODEV:
            if iodev != self.iodev:
                print(f" - {iodev!s}")
                self.p.iodev_set_usage(iodev, 0)

        print("Doing essential MMIO remaps...")
        self.map_essential()

        print("Updating page tables...")
        self.pt_update()

        adt_blob = self.adt.build()
        print(f"Uploading ADT (0x{len(adt_blob):x} bytes)...")
        self.iface.writemem(self.adt_base, adt_blob)

        print("Improving logo...")
        self.p.fb_improve_logo()

        print("Shutting down framebuffer...")
        self.p.fb_shutdown(True)

        print("Enabling SPRR...")
        self.u.msr(SPRR_CONFIG_EL1, 1)

        print("Enabling GXF...")
        self.u.msr(GXF_CONFIG_EL1, 1)

        print(f"Jumping to entrypoint at 0x{self.entry:x}")

        self.iface.dev.timeout = None
        self.default_sigint = signal.signal(signal.SIGINT, self._handle_sigint)

        set_sigquit_stackdump_handler()

        if self.wdt_cpu is not None:
            self.p.hv_wdt_start(self.wdt_cpu)
        # Does not return

        self.started = True
        for cpu_node in list(self.adt["cpus"]):
            if cpu_node.state == "running":
                break
        self.started_cpus[cpu_node.cpu_id] = (getattr(cpu_node, "die_id", 0), cpu_node.cluster_id, cpu_node.cpu_id)
        self.sysreg[cpu_node.cpu_id] = {}
        self.p.hv_start(self.entry, self.guest_base + self.bootargs_off)

from .. import trace



# SPDX-License-Identifier: MIT
from ...utils import *
from contextlib import contextmanager

PPL_MAGIC = 0x4b1d000000000002

class GFXHandoffStruct(RegMap):
    MAGIC_AP    = 0x0, Register64
    MAGIC_FW    = 0x8, Register64

    LOCK_AP     = 0x10, Register8
    LOCK_FW     = 0x11, Register8
    TURN        = 0x14, Register32

    CUR_CTX     = 0x18, Register32

    FLUSH_STATE = irange(0x20, 0x41, 0x18), Register64
    FLUSH_ADDR  = irange(0x28, 0x41, 0x18), Register64
    FLUSH_SIZE  = irange(0x30, 0x41, 0x18), Register64

    UNK2        = 0x638, Register8
    UNK3        = 0x640, Register64

class GFXHandoff:
    def __init__(self, u):
        self.u = u
        self.sgx_dev = self.u.adt["/arm-io/sgx"]
        self.base = self.sgx_dev.gfx_handoff_base
        self.reg = GFXHandoffStruct(u, self.base)
        self.is_locked = False
        self.initialized = False

    @contextmanager
    def lock(self):
        """Dekker's algorithm lock"""
        assert not self.is_locked

        # Note: This *absolutely* needs barriers everywhere.
        # Those are implicit in proxyclient for every operation.

        self.reg.LOCK_AP.val = 1
        while self.reg.LOCK_FW.val != 0:
            if self.reg.TURN != 0:
                self.reg.LOCK_AP = 0
                while self.reg.TURN != 0:
                    pass
                self.reg.LOCK_AP = 1

        self.is_locked = True
        try:
            yield
        finally:
            self.reg.TURN.val = 1
            self.reg.LOCK_AP.val = 0
            self.is_locked = False

    def initialize(self):
        if self.initialized:
            return

        print("[Handoff] Initializing...")

        self.reg.MAGIC_AP.val = PPL_MAGIC
        self.reg.UNK = 0xffffffff
        self.reg.UNK3 = 0

        with self.lock():
            print("[Handoff] Waiting for FW PPL init...")
            while self.reg.MAGIC_FW.val != PPL_MAGIC:
                pass

        for i in range(0x41):
            self.reg.FLUSH_STATE[i].val = 0
            self.reg.FLUSH_ADDR[i].val = 0
            self.reg.FLUSH_SIZE[i].val = 0

        self.initialized = True
        print("[Handoff] Initialized!")

    # The order here is:
    # - Remap memory as shared
    # - TLBI
    # - prepare_cacheflush()
    # - issue FWCtl request
    # - wait for completion (ring or wait_cacheflush?)
    # - Unmap memory
    # - TLBI
    # - complete_cacheflush()
    def prepare_cacheflush(self, base, size, context=0x40):
        assert self.reg.FLUSH_STATE[context].val == 0

        self.reg.FLUSH_ADDR[context].val = base
        self.reg.FLUSH_SIZE[context].val = size
        self.reg.FLUSH_STATE[context].val = 1

    def wait_cacheflush(self, context=0x40):
        while self.reg.FLUSH_STATE[context].val == 1:
            pass

    def complete_cacheflush(self, context=0x40):
        assert self.reg.FLUSH_STATE[context].val == 2
        self.reg.FLUSH_STATE[context].val = 0

    # probably not necessary?
    # order is:
    # - Remap memory as shared
    # - (no TLBI?)
    # - prepare_unmap()
    # - unmap
    # - TLBI
    # - complete_unmap()
    def prepare_unmap(self, base, size, context):
        assert self.reg.FLUSH_STATE[context].val == 0
        self.reg.FLUSH_ADDR[context].val = 0xdead000000000000 | (base & 0xffffffffffff)
        self.reg.FLUSH_SIZE[context].val = size
        self.reg.FLUSH_STATE[context].val = 2

    def complete_unmap(self, context):
        assert self.reg.FLUSH_STATE[context].val == 2
        self.reg.FLUSH_STATE[context].val = 0



# SPDX-License-Identifier: MIT

import struct
from construct import *
from ..constructutils import *
from ..utils import *

class HIDDescriptor(ConstructClass):
    subcon = Struct(
        "descriptor" / HexDump(GreedyBytes)
    )

class GPIOInit(ConstructClass):
    subcon = Struct(
        "unk1" / Int16ul,
        "gpio_id"/ Int16ul,
        "gpio_name" / PaddedString(32, "ascii")
    )

class InitBlock(ConstructClass):
    subcon = Struct(
        "type" / Int16ul,
        "length" / Int16ul,
        "payload" / FixedSized(this.length,
                   Switch(this.type, {
                       0: HIDDescriptor,
                       1: GPIOInit,
                       2: GreedyBytes, # Unknown 6 bytes terminator
                       7: GreedyBytes, # Device name
                   }, default=GreedyBytes))
    )

class InitMsg(ConstructClass):
    subcon = Struct(
        "msg_type" / Const(0xf0, Int8ul),
        "msg_subtype" / Const(0x01, Int8ul),
        "unk" / Const(0x00, Int8ul),
        "device_id" / Int8ul,
        "device_name" / PaddedString(16, "ascii"),
        "more_packets" / Int16ul,
        "msg" / RepeatUntil(lambda obj, lst, ctx: lst[-1].type == 2, InitBlock)
    )

class DeviceReadyMsg(ConstructClass):
    subcon = Struct(
        "msg_type" / Const(0xf1, Int8ul),
        "device_id" / Int8ul,
        "unk" / Int16ul
    )

class GPIORequestMsg(ConstructClass):
    subcon = Struct(
        "msg_type" / Const(0xa0, Int8ul),
        "device_id" / Int8ul,
        "gpio_num" / Int8ul,
        "cmd" / Int16ul,
        "args" / HexDump(GreedyBytes)
    )

NotificationMsg = Select(
    DeviceReadyMsg,
    InitMsg,
    GPIORequestMsg,
    HexDump(GreedyBytes),
)

class UnkDeviceControlMsg(ConstructClass):
    subcon = Struct(
        "command" / Int8ul,
        "args" / HexDump(GreedyBytes),
    )

class DeviceEnableMsg(ConstructClass):
    subcon = Struct(
        "command" / Const(0xb4, Int8ul),
        "device_id" / Int8ul,
    )

class DeviceResetMsg(ConstructClass):
    subcon = Struct(
        "command" / Const(0x40, Int8ul),
        "unk1" / Int8ul,
        "device_id" / Int8ul,
        "state" / Int8ul,
    )

class InitBufMsg(ConstructClass):
    subcon = Struct(
        "command" / Const(0x91, Int8ul),
        "unk1" / Int8ul,
        "unk2" / Int8ul,
        "buf_addr" / Int64ul,
        "buf_size" / Int32ul,
    )

class InitAFEMsg(ConstructClass):
    subcon = Struct(
        "command" / Const(0x95, Int8ul),
        "unk1" / Int8ul,
        "unk2" / Int8ul,
        "iface" / Int8ul,
        "buf_addr" / Int64ul,
        "buf_size" / Int32ul,
    )

class UnkMsgC1(ConstructClass):
    subcon = Struct(
        "command" / Const(0xc1, Int8ul),
        "unk1" / Int8ul,
    )

class GPIOAckMsg(ConstructClass):
    subcon = Struct(
        "command" / Const(0xa1, Int8ul),
        "unk" / Int32ul,
        "msg" / GPIORequestMsg,
    )

DeviceControlMsg = Select(
    DeviceEnableMsg,
    DeviceResetMsg,
    InitAFEMsg,
    InitBufMsg,
    UnkMsgC1,
    UnkDeviceControlMsg
)

class DeviceControlAck(ConstructClass):
    subcon = Struct(
        "command" / Int8ul
    )

class MessageHeader(ConstructClass):
    subcon = Struct(
        "flags" / Int16ul,
        "length" / Int16ul,
        "retcode" / Int32ul,
    )

class TXMessage(ConstructClass):
    subcon = Struct(
        "hdr" / MessageHeader,
        "msg" / FixedSized(this.hdr.length,
                           Switch(this.hdr.flags, {
                               0x40: HexDump(GreedyBytes),
                               0x80: DeviceControlMsg,
                               0x81: Int8ul,
                           }))
    )

    def __init__(self):
        self.hdr = MessageHeader()

class RXMessage(ConstructClass):
    subcon = Struct(
        "hdr" / MessageHeader,
        "msg" / FixedSized(this.hdr.length, HexDump(GreedyBytes)),
    )

class MTPInterface:
    def __init__(self, proto, iface):
        self.proto = proto
        self.iface = iface
        self.tx_seq = 0
        self.initialized = False
        self.gpios = {}

    def send(self, msg):
        self.proto.send(self.iface, self.tx_seq & 0xff, msg)
        self.tx_seq += 1

    def get_report(self, idx):
        msg = TXMessage()
        msg.hdr.flags = 0x81
        msg.hdr.length = 1
        msg.hdr.retcode = 0
        msg.msg = idx
        self.send(msg.build())

    def packet(self, pkt):
        self.log(f"RX: {pkt.hex()}")

    def log(self, s):
        self.proto.log(f"[{self.NAME}] " + s)

    def initialize(self):
        self.proto.comm.enable_device(self.iface)

    def report(self, msg):
        self.log(f"report: {msg.hex()}")

    def ack(self, msg):
        self.log(f"ack: {msg.hex()}")

    def unk(self, msg):
        self.log(f"unk: {msg.hex()}")

    def packet(self, pkt):
        msg = RXMessage.parse(pkt)
        mtype = msg.hdr.flags
        #self.log(f"FL:{msg.hdr.flag    s:04x} unk:{msg.hdr.unk:08x}")
        if mtype == 0x00:
            self.report(msg.msg)
        elif mtype == 0x80:
            self.ack(msg.hdr.retcode, msg.msg)
        elif mtype == 0x81:
            self.log(f"REPORT")
            chexdump(msg.msg, print_fn=self.log)
        elif mtype == 0x40:
            self.unk(msg.msg)

    def __str__(self):
        return f"{self.iface}/{self.NAME}"


class MTPCommInterface(MTPInterface):
    NAME = "comm"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_cmd = None
        self.gpios = {}


    def device_control(self, dcmsg):
        while self.last_cmd is not None:
            self.proto.work()
        msg = TXMessage()
        msg.hdr.flags = 0x80
        msg.hdr.length = len(dcmsg.build())
        msg.hdr.retcode = 0
        msg.msg = dcmsg
        #self.log(f"Send device control {dcmsg}")
        self.last_cmd = dcmsg.command
        self.send(msg.build())
        while self.last_cmd is not None:
            self.proto.work()

    def enable_device(self, iface):
        msg = DeviceEnableMsg()
        msg.device_id = iface
        self.device_control(msg)

    def report(self, msg):
        msg = NotificationMsg.parse(msg)

        if isinstance(msg, DeviceReadyMsg):
            iface = self.proto.iface[msg.device_id]
            iface.initialized = True
            self.log(f"{iface}: init complete")
        elif isinstance(msg, InitMsg):
            iface = self.proto.get_interface(msg.device_id, msg.device_name)
            for blk in msg.msg:
                if isinstance(blk.payload, HIDDescriptor):
                    self.log(f"Got HID descriptor for {iface}:")
                    iface.descriptor = blk.payload.descriptor
                    self.log(hexdump(iface.descriptor))
                elif isinstance(blk.payload, GPIOInit):
                    self.log(f"GPIO Init: {blk.payload}")
                    prop = getattr(self.proto.node[msg.device_name],
                                   f"function-{blk.payload.gpio_name}".replace("-", "_"))
                    key = struct.pack(">I", prop.args[0]).decode("ascii")
                    val = prop.args[1]
                    self.log(f"GPIO key: {key}")
                    self.gpios[(msg.device_id, blk.payload.gpio_id)] = key, val
            if not msg.more_packets:
                iface.initialize()
        elif isinstance(msg, GPIORequestMsg):
            self.log(f"GPIO request: {msg}")
            smcep = self.proto.smc.epmap[0x20]
            key, val = self.gpios[(msg.device_id, msg.gpio_num)]
            if msg.cmd == 3:
                smcep.write32(key, val | 1)
                smcep.write32(key, val)

            ackmsg = GPIOAckMsg()
            ackmsg.unk = 0
            ackmsg.msg = msg
            self.device_control(ackmsg)
        else:
            self.log("Unknown message!")
            print(msg)

    def ack(self, retcode, msg):
        msg = DeviceControlAck.parse(msg)
        self.log(f"Got ACK for {msg.command:#x}: {retcode:08x}")
        assert msg.command == self.last_cmd
        self.last_cmd = None

    def init_afe(self, iface, data):
        paddr, dva = self.proto.mtp.ioalloc(len(data))
        self.proto.u.iface.writemem(paddr, data)

        afemsg = InitAFEMsg()
        afemsg.unk1 = 2
        afemsg.unk2 = 0
        afemsg.iface = iface
        afemsg.buf_addr = dva
        afemsg.buf_size = len(data)
        self.device_control(afemsg)

    def device_reset(self, iface, unk1, state):
        self.log(f"device_reset({iface}, {unk1}, {state})")
        rmsg = DeviceResetMsg()
        rmsg.device_id = iface
        rmsg.unk1 = unk1
        rmsg.state = state
        self.device_control(rmsg)

class MTPHIDInterface(MTPInterface):
    pass

class MTPMultitouchInterface(MTPHIDInterface):
    NAME = "multi-touch"

    def initialize(self):
        super().initialize()

        #data = open("afe.bin", "rb").read()
        #self.proto.comm.init_afe(self.iface, data)
        #self.proto.comm.device_reset(self.iface, 1, 0)
        #self.proto.comm.device_reset(self.iface, 1, 2)

class MTPKeyboardInterface(MTPHIDInterface):
    NAME = "keyboard"

class MTPSTMInterface(MTPHIDInterface):
    NAME = "stm"

class MTPActuatorInterface(MTPHIDInterface):
    NAME = "actuator"

class MTPTPAccelInterface(MTPHIDInterface):
    NAME = "tp_accel"

class MTPProtocol:
    INTERFACES = [
        MTPCommInterface,
        MTPMultitouchInterface,
        MTPKeyboardInterface,
        MTPSTMInterface,
        MTPActuatorInterface,
        MTPTPAccelInterface,
    ]

    def __init__(self, u, node, mtp, dockchannel, smc):
        self.node = node
        self.smc = smc
        self.u = u
        self.mtp = mtp
        self.dockchannel = dockchannel
        self.iface = {}

        # Add initial comm interface
        self.get_interface(0, "comm")

    def get_interface(self, iface, name):
        if iface in self.iface:
            return self.iface[iface]

        for cls in self.INTERFACES:
            if cls.NAME == name:
                break
        else:
            self.log(f"Unknown interface name {name}")
            return None
        obj = cls(self, iface)
        self.iface[iface] = obj
        setattr(self, name.replace("-", "_"), obj)
        return obj

    def checksum(self, d):
        assert len(d) % 4 == 0
        c = len(d) // 4
        return 0xffffffff - sum(struct.unpack(f"<{c}I", d)) & 0xffffffff

    def read_pkt(self):
        self.mtp.work_pending()
        hdr = self.dockchannel.read(8)
        hlen, mtype, size, ctr, devid, pad = struct.unpack("<BBHBBH", hdr)
        #self.log(f"<L:{hlen} T:{mtype:02x} S:{size:04x} D:{devid}")
        assert hlen == 8
        #assert mtype == 0x12
        data = self.dockchannel.read(size)
        checksum = struct.unpack("<I", self.dockchannel.read(4))[0]
        expect = self.checksum(hdr + data)
        if expect != checksum:
            self.log(f"Checksum error: expected {expect:08x}, got {checksum:08x}")
        return devid, data

    def send(self, iface, seq, msg):
        if len(msg) % 4:
            msg += bytes(4 - len(msg) % 4)
        hdr = struct.pack("<BBHBBH", 8, 0x11, len(msg), seq, iface, 0)
        checksum = self.checksum(hdr + msg)
        pkt = hdr + msg + struct.pack("<I", checksum)
        self.dockchannel.write(pkt)
        self.mtp.work_pending()

    def work_pending(self):
        self.mtp.work_pending()
        while self.dockchannel.rx_count != 0:
            self.work()
        self.mtp.work_pending()

    def work(self):
        devid, pkt = self.read_pkt()
        self.iface[devid].packet(pkt)

    def wait_init(self, name):
        self.log(f"Waiting for {name}...")
        while not hasattr(self, name) or not getattr(self, name).initialized:
            self.work()

    def log(self, m):
        print("[MTP]" + m)



import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm
from m1n1.shell import run_shell
from m1n1.gpiola import GPIOLogicAnalyzer
from m1n1.hw.spi import *

p.smp_start_secondaries()

spi = u.adt["arm-io/spi2"].get_reg(0)[0] + 0x8000
regs = SPIRegs(u, spi)

aic = u.adt["arm-io/aic"].get_reg(0)[0]

gpio = u.adt["arm-io/gpio0"].get_reg(0)[0]
count = u.adt["arm-io/gpio0"].getprop("#gpio-pins")

pins={}
for i in range(150, 150+32):
    pins[f"pin{i}"] = i

m = GPIOLogicAnalyzer(u, "arm-io/gpio0",
                      pins=pins,
                      regs={"a": gpio},
                      div=1,
                      cpu=1,
                      on_pin_change=True,
                      on_reg_change=False)

regs.CTRL.val = 0xc
regs.PIN.val = 0x2
regs.CFG.val = 0x20 | (1<<15) | 6
regs.CFG.val = 0x20 | (1<<15) | 4
regs.CFG.val = 0x20 | (1<<15) | 2
regs.CFG.val = 0x20 | (3<<15) | 0

m.regs = {}

m.start(30000000, bufsize=0x80000)

regs.STATUS.val = 0xffffffff
regs.IF_XFER.val = 0xffffffff
regs.IF_FIFO.val = 0xffffffff

regs.CLKDIV.val = 0xfff
regs.INTER_DELAY.val = 0x1000

regs.SHIFTCFG.val = 0x21fcf7

regs.PIN.val = 0x2
print("pinconfig", hex(regs.PINCFG.val))
regs.PINCFG.val = 0x100
#regs.PINCONFIG.val = 0x2-7
print("pinconfig", hex(regs.PINCFG.val))
print("shiftconfig", hex(regs.SHIFTCFG.val))

p.write32(spi + 0x3c, 0xffffffff)

regs.PINCFG.val = 0x002
regs.PINCFG.val = 0x200

#p.write32(0x28e0380bc, 0x80100000)
#p.write32(0x28e0380c4, 0x80100000)

data = b"\xff\xff\xff\xff\x00\x00\xff\xff"

for i in range(2):
    for j in data:
        regs.TXDATA.val = j | 0xffffff00
    regs.RXCNT.val = len(data)
    regs.TXCNT.val = len(data)

    regs.STATUS.val = 0xffffffff
    regs.IF_XFER.val = 0xffffffff
    regs.IF_FIFO.val = 0xffffffff

    regs.PIN.val = 0x0
    regs.CTRL.val = 0x1
    #regs.TXDATA.val = 0xff
    #regs.TXDATA.val = 0xff

    i = 0
    while regs.TXCNT.val != 0:
        print(f"{regs.TXCNT.val:#x} {regs.FIFOSTAT.reg} {regs.STATUS.val:#x} {regs.IF_FIFO.val:#x} {p.read32(spi + 0x134):#x}")
        regs.STATUS.val = 0xffffffff
        regs.IF_XFER.val = 0xffffffff
        regs.IF_FIFO.val = 0xffffffff
        #regs.CTRL.val = 0x0
        #time.sleep(0.1)
        #regs.CTRL.val = 0x1[
        print(hex(i))
        #p.write32(spi + i, 0xffffffff)
        #p.write32(spi + i, 0)
        i += 4
        if i > 0x100:
            break
    time.sleep(0.001)
    print(f"{regs.RXCNT.val:#x} {regs.FIFOSTAT.reg} {regs.STATUS.val:#x} {regs.IF_FIFO.val:#x}")
    regs.STATUS.val = 0xffffffff
    regs.IF_XFER.val = 0xffffffff
    regs.IF_FIFO.val = 0xffffffff

    mon.poll()
    
    while regs.FIFOSTAT.reg.LEVEL_RX:
        print("RX", hex(regs.RXDATA.val))

    regs.CTRL.val = 0

m.complete()
m.show()

#run_shell(globals(), msg="Have fun!")





# SPDX-License-Identifier: MIT
import struct

from construct import *
from m1n1.utils import *
from m1n1.proxyutils import *
from m1n1.constructutils import *
from m1n1.trace.asc import ASCTracer, EP, EPState, msg, msg_log, DIR
from m1n1.trace.dockchannel import DockChannelTracer
from m1n1.trace.dart import DARTTracer
from m1n1.fw.mtp import *

class MTPTracer(ASCTracer):
    def handle_msg(self, direction, r0, r1):
        super().handle_msg(direction, r0, r1)

mtp_tracer = MTPTracer(hv, "/arm-io/mtp", verbose=1)
mtp_tracer.start()

dart_tracer = DARTTracer(hv, "/arm-io/dart-mtp", verbose=1)
dart_tracer.start()

trace_device("/arm-io/dart-mtp", True)

DockChannelTracer = DockChannelTracer._reloadcls()

mon = RegMonitor(hv.u, ascii=True, bufsize=0x400000)

class StreamState:
    def __init__(self):
        self.buf = bytes()

class MTPStream:
    def __init__(self, tracer, name, state):
        self.tracer = tracer
        self.name = name
        self.state = state

    def put(self, d):
        buf = self.state.buf + d

        while buf:
            if len(buf) < 8:
                self.state.buf = buf
                return

            hlen, mtype, size, ctr, devid, pad = struct.unpack("<BBHBBH", buf[:8])
            assert pad == 0
            if hlen != 8:
                self.tracer.log(f"Bad hlen 0x{hlen:2x}, skipping a byte...")
                self.state.buf = buf[1:]
                return


            need = 8 + size + 4
            if len(buf) < need:
                self.state.buf = buf
                return

            payload = buf[8:8 + size]
            self.packet(mtype, devid, ctr, payload)
            buf = buf[need:]

        self.state.buf = buf

    def packet(self, mtype, devid, ctr, pkt):
        self.tracer.packet(mtype, devid, ctr, pkt, dir=self.name)

    def log(self, msg):
        self.tracer.log(f"{self.name} " + msg)

class MTPChannelTracer(DockChannelTracer):
    def init_state(self):
        self.state.rx = StreamState()
        self.state.tx = StreamState()
        self.state.rx_mem = StreamState()
        self.state.buf = None
        self.state.buf_size = None
        self.state.rptr = 0

    def start(self, dart=None):
        super().start()
        self.dart = dart
        self.rx_stream = MTPStream(self, "<", self.state.rx)
        self.tx_stream = MTPStream(self, ">", self.state.tx)
        self.rx_mem_stream = MTPStream(self, "<", self.state.rx_mem)
        self.init_mon()

    def tx(self, d):
        self.tx_stream.put(d)

    def rx(self, d):
        self.rx_stream.put(d)

    def init_mon(self):
        pass
        #if self.state.buf is not None:
            #addr, size = self.dart.iotranslate(1, self.state.buf, self.state.buf_size)[0]
            #size = align_up(size, 4)
            #mon.add(addr, size)

    def poll_ring(self):
        wptr = struct.unpack("<I", self.dart.ioread(1, self.state.buf, 4))[0]
        rptr = self.state.rptr
        if wptr < rptr:
            size = self.state.buf_size - 8 - rptr
            d = self.dart.ioread(1, self.state.buf + 8 + rptr, size)
            self.rx_mem_stream.put(d)
            self.state.rptr = rptr = 0

        size = wptr - rptr
        d = self.dart.ioread(1, self.state.buf + 8 + rptr, size)
        self.rx_mem_stream.put(d)
        self.state.rptr = rptr + size

    def packet(self, mcode, devid, ctr, data, dir):
        #mon.poll()
        chexdump(data, print_fn=self.log)

        if data == b"":
            self.poll_ring()
            msg = "<null>"
        elif dir == ">":
            msg = TXMessage.parse(data)
        elif dir == "<":
            msg = RXMessage.parse(data)
            if devid == 0:
                if (msg.hdr.flags & 0xc0) == 0x00:
                    msg.msg = NotificationMsg.parse(msg.msg)
                elif (msg.hdr.flags & 0xc0) == 0x80:
                    msg.msg = DeviceControlAck.parse(msg.msg)
        else:
            assert False

        try:
            mtype = msg.msg.name
        except:
            mtype = None

        if mtype == "InitAFEMsg":
            afe = self.dart.ioread(1, msg.msg.buf_addr, msg.msg.buf_size)
            chexdump(afe, print_fn=self.log)
            open("afe.bin", "wb").write(afe)
        elif mtype == "InitBufMsg":
            self.state.buf = msg.msg.buf_addr
            self.state.buf_size = msg.msg.buf_size
            self.init_mon()

        self.log(f"{dir} Type {mcode:02x} Dev {devid} #{ctr} {msg!s}")

hid_tracer = MTPChannelTracer(hv, "/arm-io/dockchannel-mtp", verbose=3)
hid_tracer.start(dart_tracer.dart)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import atexit, sys

from m1n1.setup import *
from m1n1.constructutils import Ver
from m1n1.utils import *

Ver.set_version(u)

from m1n1.agx import AGX
from m1n1.agx.render import *

from m1n1 import asm

from m1n1.gpiola import GPIOLogicAnalyzer

analyzer_cpu = 1

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")
p.smp_start_secondaries()
p.mmu_init_secondary(analyzer_cpu)
iface.dev.timeout = 42

agx = AGX(u)

mon = RegMonitor(u, ascii=True, bufsize=0x8000000)
agx.mon = mon

sgx = agx.sgx_dev

atexit.register(p.reboot)
agx.start()

mon.poll()
agx.poll_objects()

ctx = GPUContext(agx)
ctx.bind(63)

f = GPUFrame(ctx, sys.argv[1], track=False)

r = GPURenderer(ctx, 128, bm_slot=0x10, queue=1)

dep_stamp = agx.kobj.new(StampCounter, name="Dep stamp")
dep_stamp.value = 0x100
dep_stamp.push()

#r.submit(f.cmdbuf, (dep_stamp._addr, 0x200, 0x10))
r.submit(f.cmdbuf)
r.submit(f.cmdbuf)

def t(addr):
    paddr = agx.uat.iotranslate(0, addr, 4)[0][0]
    if paddr is None:
        raise Exception(f"Failed to iotranslate {addr:#x}")
    return paddr

regs = {
    "ta_cmds":  t(agx.initdata.regionB.stats_ta.addrof("total_cmds")),
    "ta_ts":    t(agx.initdata.regionB.stats_ta.stats.addrof("unk_timestamp")),
}

pend_base = agx.initdata.regionC.addrof("pending_stamps")
for i in range(5):
    regs[f"st{i}_info"] = t(pend_base + i*8)
    regs[f"st{i}_val"] = t(pend_base + i*8 + 4)

for i in range(4):
    regs[f"ta{i}_cq"] = t(agx.initdata.regionB.stats_ta.stats.queues[i].addrof("cur_cmdqueue"))

regs.update({
    #"pwr_status": t(agx.initdata.regionB.hwdata_a.addrof("pwr_status")),
    #"pstate": t(agx.initdata.regionB.hwdata_a.addrof("cur_pstate")),
    #"temp_c": t(agx.initdata.regionB.hwdata_a.addrof("temp_c")),
    #"pwr_mw": t(agx.initdata.regionB.hwdata_a.addrof("avg_power_mw")),
    #"pwr_ts": t(agx.initdata.regionB.hwdata_a.addrof("update_ts")),

    #"unk_10": t(agx.initdata.regionB.hwdata_a.addrof("unk_10")),
    #"unk_14": t(agx.initdata.regionB.hwdata_a.addrof("unk_14")),
    #"actual_pstate": t(agx.initdata.regionB.hwdata_a.addrof("actual_pstate")),
    #"tgt_pstate": t(agx.initdata.regionB.hwdata_a.addrof("tgt_pstate")),
    #"unk_40": t(agx.initdata.regionB.hwdata_a.addrof("unk_40")),
    #"unk_44": t(agx.initdata.regionB.hwdata_a.addrof("unk_44")),
    #"unk_48": t(agx.initdata.regionB.hwdata_a.addrof("unk_48")),
    #"freq_mhz": t(agx.initdata.regionB.hwdata_a.addrof("freq_mhz")),

    #"unk_748.0": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")),
    #"unk_748.1": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+4),
    #"unk_748.2": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+8),
    #"unk_748.3": t(agx.initdata.regionB.hwdata_a.addrof("unk_748")+12),
    #"use_percent": t(agx.initdata.regionB.hwdata_a.addrof("use_percent")),
    #"unk_83c": t(agx.initdata.regionB.hwdata_a.addrof("unk_83c")),
    #"freq_with_off": t(agx.initdata.regionB.hwdata_a.addrof("freq_with_off")),
    #"unk_ba0": t(agx.initdata.regionB.hwdata_a.addrof("unk_ba0")),
    #"unk_bb0": t(agx.initdata.regionB.hwdata_a.addrof("unk_bb0")),
    #"unk_c44": t(agx.initdata.regionB.hwdata_a.addrof("unk_c44")),
    #"unk_c58": t(agx.initdata.regionB.hwdata_a.addrof("unk_c58")),

    #"unk_3ca0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3ca0")),
    #"unk_3ca8": t(agx.initdata.regionB.hwdata_a.addrof("unk_3ca8")),
    #"unk_3cb0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3cb0")),
    #"ts_last_idle": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_idle")),
    #"ts_last_poweron": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_poweron")),
    #"ts_last_poweroff": t(agx.initdata.regionB.hwdata_a.addrof("ts_last_poweroff")),
    #"unk_3cd0": t(agx.initdata.regionB.hwdata_a.addrof("unk_3cd0")),

    "halt_count":  t(agx.initdata.fw_status.addrof("halt_count")),
    "halted":  t(agx.initdata.fw_status.addrof("halted")),
    "resume":  t(agx.initdata.fw_status.addrof("resume")),
    "unk_40":  t(agx.initdata.fw_status.addrof("unk_40")),
    "unk_ctr":  t(agx.initdata.fw_status.addrof("unk_ctr")),
    "unk_60":  t(agx.initdata.fw_status.addrof("unk_60")),
    "unk_70":  t(agx.initdata.fw_status.addrof("unk_70")),
    "c_118c0":  t(agx.initdata.regionC._addr + 0x118c0),
    "c_118c4":  t(agx.initdata.regionC._addr + 0x118c4),
    "c_118c8":  t(agx.initdata.regionC._addr + 0x118c8),
    "c_118cc":  t(agx.initdata.regionC._addr + 0x118cc),
    "c_118d0":  t(agx.initdata.regionC._addr + 0x118d0),
    "c_118d4":  t(agx.initdata.regionC._addr + 0x118d4),
    "c_118d8":  t(agx.initdata.regionC._addr + 0x118d8),
    "c_118dc":  t(agx.initdata.regionC._addr + 0x118dc),
    "3d_cmds":  t(agx.initdata.regionB.stats_3d.addrof("total_cmds")),
    #"3d_tvb_oflws_1":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_1")),
    #"3d_tvb_oflws_2":   t(agx.initdata.regionB.stats_3d.stats.addrof("tvb_overflows_2")),
    "3d_cur_stamp_id":  t(agx.initdata.regionB.stats_3d.stats.addrof("cur_stamp_id")),
    "3d_ts":    t(agx.initdata.regionB.stats_3d.stats.addrof("unk_timestamp")),
    #"3d_cur_stamp_id":  t(agx.initdata.regionB.stats_3d.stats.addrof("cur_stamp_id")),
})

for i in range(4):
    regs[f"3d{i}_cq"] = t(agx.initdata.regionB.stats_3d.stats.queues[i].addrof("cur_cmdqueue"))


i = 0
regs.update({
    f"r{i}_3d_done":  t(r.wq_3d.info.pointers.addrof("gpu_doneptr")),
    #f"r{i}_3d_rptr":  t(r.wq_3d.info.pointers.addrof("gpu_rptr")),
    f"r{i}_3d_busy":  t(r.wq_3d.info.addrof("busy")),
    #f"r{i}_3d_blk":   t(r.wq_3d.info.addrof("blocked_on_barrier")),
    #f"r{i}_3d_2c":    t(r.wq_3d.info.addrof("unk_2c")),
    #f"r{i}_3d_54":    t(r.wq_3d.info.addrof("unk_54")),

    f"r{i}_ta_done":  t(r.wq_ta.info.pointers.addrof("gpu_doneptr")),
    #f"r{i}_ta_rptr":  t(r.wq_ta.info.pointers.addrof("gpu_rptr")),
    f"r{i}_ta_busy":  t(r.wq_ta.info.addrof("busy")),
    #f"r{i}_ta_blk":   t(r.wq_ta.info.addrof("blocked_on_barrier")),
    #f"r{i}_ta_2c":    t(r.wq_ta.info.addrof("unk_2c")),
    #f"r{i}_ta_54":    t(r.wq_ta.info.addrof("unk_54")),
    f"r{i}_ta_stamp1": t(r.stamp_ta1._addr),
    f"r{i}_ta_stamp2":t(r.stamp_ta2._addr),
    f"r{i}_3d_stamp1": t(r.stamp_3d1._addr),
    f"r{i}_3d_stamp2":t(r.stamp_3d2._addr),

    f"r{i}_ev_cnt":t(r.event_control.event_count._addr),
    f"r{i}_ev_cur":t(r.event_control.addrof("cur_count")),
    f"r{i}_ev_10":t(r.event_control.addrof("unk_10")),
})

div=4
ticks = 24000000 // div * 25

la = GPIOLogicAnalyzer(u, regs=regs, cpu=analyzer_cpu, div=div)

print("Queues:")
print(f"  TA: {r.wq_ta.info._addr:#x} (stamp {r.work[0].ev_ta.id})")
#print(r.wq_ta.info)
print(f"  3D: {r.wq_3d.info._addr:#x} (stamp {r.work[0].ev_3d.id})")
#print(r.wq_3d.info)

print("==========================================")
print("## Run")
print("==========================================")

la.start(ticks, bufsize=0x8000000)

t = time.time()

buf = agx.kobj.new_buf(0x1000, "foo")
buf.val = b"A" * 0x1000
buf.push()
agx.uat.flush_dirty()

try:
    r.run()

    #for a in range(8):
        #for b in range(8):
            #agx.ch.devctrl.dc_1e(a, b)

    #agx.uat.flush_dirty()
    #agx.ch.devctrl.write32(dep_stamp._addr, 0x200)

    #data = struct.pack("<QQQQQI", 0xaaaa, buf._addr, 0xbbbb, 0, 0, 0)


    #agx.poll_objects()
    #mon.poll()
    #agx.ch.devctrl.send_foo(9, data)
    #agx.ch.devctrl.dc_09(0xaaaa, buf._addr, 0xbbbb)

    #for i in range(0x28, 0xff):
        #print(hex(i))
        #data = struct.pack("<QQQQQI", dep_stamp._addr, 0x10_00000200, buf._addr, 0x12_00000010, 0x13_00000010, 0x10)
        #agx.ch.devctrl.send_foo(i, data)
        #agx.asc.work()
        #time.sleep(0.1)
        #agx.poll_objects()
        #mon.poll()

    #time.sleep(0.1)
    #agx.asc.work()

    #chexdump(buf.pull().val)

    agx.kick_firmware()

    while not r.ev_3d.fired:
        agx.asc.work()
        agx.poll_channels()
        print("==========================================")
        #agx.poll_objects()
        #mon.poll()
        agx.kick_firmware()
        if time.time() > t + 2:
            raise Exception("Timeout")
    r.wait()

finally:

    dep_stamp.pull()
    print(f"Stamp value: {dep_stamp.value:#x}")

    #agx.poll_objects()
    #mon.poll()

    la.complete()
    la.show()

time.sleep(2)




#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.find_regs import *
from m1n1 import asm

p.iodev_set_usage(IODEV.FB, 0)

if u.mrs(SPRR_CONFIG_EL1):
    u.msr(GXF_CONFIG_EL12, 0)
    u.msr(SPRR_CONFIG_EL12, 0)
    u.msr(GXF_CONFIG_EL1, 0)
    u.msr(SPRR_CONFIG_EL1, 0)

# Set up HCR_EL2 for EL1, since we can't do it after enabling GXF
u.inst("nop", call="el1")

all_regs = set()
for reg in [SPRR_CONFIG_EL1, GXF_CONFIG_EL1, SPRR_CONFIG_EL12, GXF_CONFIG_EL12]:
    old_regs = set(find_regs(u, values=False))
    u.msr(reg, 1)
    el2_items = set(find_regs(u))
    el2_vals = dict(el2_items)
    new_regs = set(k for k, v in el2_items)

    all_regs = all_regs.union(new_regs)

    diff_regs = new_regs - old_regs

    print(reg)
    for r in sorted(diff_regs):
        print("  %s --> %lx" % (sysreg_name(r), u.mrs(r)))

gl2_items = list(find_regs(u, regs=static_regs,call="gl2"))
gl2_vals = dict(gl2_items)
gl2_regs = set(k for k, v in gl2_items)

print("GL2")
for reg in sorted(gl2_regs - all_regs):
    print("  %s -> %lx" % (sysreg_name(reg), gl2_vals[reg]))
for reg in sorted(gl2_regs):
    if reg in el2_vals and gl2_vals[reg] != el2_vals[reg]:
        print("  ! %s %lx -> %lx" % (sysreg_name(reg), el2_vals[reg], gl2_vals[reg]))

u.msr(GXF_CONFIG_EL12, 0)
u.msr(SPRR_CONFIG_EL12, 0)
u.msr(GXF_CONFIG_EL1, 0)
u.msr(SPRR_CONFIG_EL1, 0)

gl1_items = list(find_regs(u, regs=static_regs, call="gl1"))
gl1_vals = dict(gl1_items)
gl1_regs = set(k for k, v in gl1_items)

print("GL1")
for reg in sorted(gl1_regs - all_regs):
    val = gl1_vals[reg]
    print("  %s -> %lx" % (sysreg_name(reg), val))

    cval = u.mrs(reg, call="gl1", silent=False)
    print("    cur: 0x%lx" % (cval))

    try:
        u.msr(reg, cval, call="gl1", silent=False)
    except:
        print(">RO")
        continue

    gl2_vals = dict(find_regs(u, regs=static_regs,call="gl2"))
    u.msr(reg, cval ^ 0xffff, call="gl1", silent=True)

    for r, v in find_regs(u, regs=static_regs, call="gl2"):
        if v != gl2_vals[r]:
            print("     GL2 access: %s %lx -> %lx" % (sysreg_name(r), gl2_vals[r], v))

    u.msr(reg, cval, call="gl1", silent=True)

for reg in sorted(gl1_regs):
    if reg in el2_vals and gl1_vals[reg] != el2_vals[reg]:
        print("  ! %s %lx -> %lx" % (sysreg_name(reg), el2_vals[reg], gl1_vals[reg]))



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum

__all__ = ["I2C", "I2CRegs"]


class R_MTXFIFO(Register32):
    READ    = 10        # Read (DATA=count)
    STOP    = 9         # Issue START before
    START   = 8         # Issue STOP after
    DATA    = 7, 0      # Byte to send or count

class R_MRXFIFO(Register32):
    EMPTY   = 8         # FIFO empty
    DATA    = 7, 0      # FIFO data

class R_MCNT(Register32):
    S_RXCNT = 31, 24    # Slave RX count
    S_TXCNT = 23, 16    # Slave TX count
    M_RXCNT = 15, 8     # Master RX count
    M_TXCNT = 7, 0      # Master TX count

class E_MST(IntEnum):
    IDLE    = 0
    FRD1    = 1
    FRD2    = 2
    COMMAND = 3
    START   = 4
    WRITE   = 5
    READ    = 6
    ACK     = 7
    STOP    = 8
    BAD     = 15

class E_SST(IntEnum):
    IDLE    = 0
    START   = 1
    ST_ACK  = 2
    DATA    = 3
    ACK     = 4

class R_XFSTA(Register32):
    MST     = 27, 24, E_MST # Master controller state
    SRD     = 20            # Slave read in progress
    SWR     = 19            # Slave write in progress
    SST     = 18, 16, E_SST # Slave controller state
    XFIFO   = 9, 8          # FIFO number for error
    XFCNT   = 7, 0          # Number of bytes in current xfer

class R_SADDR(Register32):
    DEB     = 31            # Enable SDA/SCL read debug
    DIR     = 30            # Direct (bitbang) mode
    ENS     = 29            # Enable slave interface
    RST_STX = 28            # Reset slave TX FIFO
    RST_SRX = 27            # Reset master RX fifo (if ^ both, controller too)
    PEN     = 26            # Promiscuous mode (slave)
    AAE     = 25            # SALT/ALTMASK enable
    SAE     = 24            # SADDR enable
    ALTMASK = 23, 16        # MASK for SALT bits
    SALT    = 15, 8         # Alt slave address
    SADDR   = 7, 0          # Slave address

class R_SMSTA(Register32):
    XIP     = 28            # Xaction in progress
    XEN     = 27            # Xaction ended
    UJF     = 26            # UnJam failure
    JMD     = 25            # Jam ocurred
    JAM     = 24            # Currently jammed
    MTO     = 23            # Master timeout
    MTA     = 22            # Master arb lost
    MTN     = 21            # Master received NACK
    MRF     = 20            # Master RX fifo full
    MRNE    = 19            # Master RX fifo not empty
    MTF     = 17            # Master TX fifo full
    MTE     = 16            # Master RX fifo empty
    STO     = 15            # Slave timeout
    STA     = 14            # Slave arb lost
    STN     = 13            # Slave received NACK
    SRF     = 12            # Slave RX fifo full
    SRNE    = 11            # Slave RX fifo not empty
    STR     = 10            # Slave transmit required
    STF     = 9             # Slave TX fifo full
    STE     = 8             # Slave TX fifo empty
    TOS     = 7             # Timeout due to slave FIFO
    TOM     = 6             # Timeout due to master FIFO
    TOE     = 5             # Slave timeout due to ext clock stretch
    DCI     = 4             # Direct clock in
    DDI     = 3             # Direct data in
    DCO     = 2             # Direct clock out
    DDO     = 1             # Direct data out
    NN      = 0             # NACK next (slave)

class R_CTL(Register32):
    MSW     = 26, 16        # Maximum slave write size
    ENABLE  = 11            # Unknown enable bit (clock sel? Apple thing)
    MRR     = 10            # Master receive FIFO reset
    MTR     = 9             # Master transmit FIFO reset
    UJM     = 8             # Enable auto unjam machine
    CLK     = 7, 0          # Clock divider

class R_STXFIFO(Register32):
    DATA    = 7, 0          # Data

class R_SRXFIFO(Register32):
    N       = 12            # NACK received after this byte
    P       = 11            # Stop received, data not valid
    S       = 10            # Start received before
    O       = 9             # Overflow (promisc only)
    E       = 8             # Empty (data not valid)
    DATA    = 7, 0          # Data

# Apple reg
class R_FIFOCTL(Register32):
    HALT    = 0             # Halt machinery

class I2CRegs(RegMap):
    MTXFIFO = 0x00, R_MTXFIFO
    MRXFIFO = 0x04, R_MRXFIFO
    MCNT    = 0x08, R_MCNT
    XFSTA   = 0x0c, R_XFSTA
    SADDR   = 0x10, R_SADDR
    SMSTA   = 0x14, R_SMSTA
    IMASK   = 0x18, R_SMSTA
    CTL     = 0x1c, R_CTL
    STXFIFO = 0x20, R_STXFIFO
    SRXFIFO = 0x20, R_SRXFIFO
    FIFOCTL = 0x44, R_FIFOCTL


class I2C:
    def __init__(self, u, adt_path):
        self.u = u
        self.p = u.proxy
        self.iface = u.iface
        self.base = u.adt[adt_path].get_reg(0)[0]
        self.regs = I2CRegs(u, self.base)
        self.devs = []

    def clear_fifos(self):
        self.regs.CTL.set(MTR=1, MRR=1)

    def clear_status(self):
        self.regs.SMSTA.val = 0xffffffff

    def _fifo_read(self, nbytes):
        read = []
        for _ in range(nbytes):
            val = self.regs.MRXFIFO.reg
            timeout = 10000
            while val.EMPTY and timeout > 0:
                val = self.regs.MRXFIFO.reg
                timeout -= 1
            if timeout == 0:
                raise Exception("timeout")
            read.append(int(val) & 0xff)
        return bytes(read)

    def _fifo_write(self, buf, stop=False):
        for no, byte in enumerate(buf):
            sending_stop = stop and no == len(buf) - 1
            self.regs.MTXFIFO.set(DATA=byte, STOP=int(sending_stop))

        if not stop:
            return

        timeout = 10000
        while not self.regs.SMSTA.reg.XEN and timeout > 0:
            timeout -= 1
        if timeout == 0:
            raise Exception("timeout")

    def write_reg(self, addr, reg, data, regaddrlen=1):
        self.clear_fifos()
        self.clear_status()

        self.regs.CTL.set(ENABLE=1, CLK=0x4)
        self.regs.MTXFIFO.set(DATA=addr << 1, START=1)
        regbytes = int.to_bytes(reg, regaddrlen, byteorder="big")
        self._fifo_write(regbytes + bytes(data), stop=True)
        self.regs.CTL.set(ENABLE=0, CLK=0x4)

    def read_reg(self, addr, reg, nbytes, regaddrlen=1):
        self.clear_fifos()
        self.clear_status()

        self.regs.CTL.set(ENABLE=1, CLK=0x4)
        self.regs.MTXFIFO.set(DATA=addr << 1, START=1)
        regbytes = int.to_bytes(reg, regaddrlen, byteorder="big")
        self._fifo_write(regbytes, stop=False)
        self.regs.MTXFIFO.set(DATA=(addr << 1) | 1, START=1)
        self.regs.MTXFIFO.set(DATA=nbytes, STOP=1, READ=1)
        data = self._fifo_read(nbytes)
        self.regs.CTL.set(ENABLE=0, CLK=0x4)
        return data

class I2CRegMapDev:
    REGMAP = None
    ADDRESSING = (0, 1)

    def __init__(self, bus, addr, name=None):
        self.bus = bus
        self.addr = addr
        self.curr_page = None
        self.name = name

        self.paged, self.regimmbytes = self.ADDRESSING
        if self.REGMAP is not None:
            self.regs = self.REGMAP(self, 0)

    @classmethod
    def from_adt(cls, bus, path):
        node = bus.u.adt[path]
        addr = node.reg[0] & 0xff
        return cls(bus, addr, node.name)

    def _switch_page(self, page):
        assert self.paged
        self.bus.write_reg(self.addr, 0, bytes([page]),
                            regaddrlen=self.regimmbytes)
        self.curr_page = page

    def _snip_regaddr(self, addr):
        pageshift = self.regimmbytes * 8
        page = addr >> pageshift
        immediate = addr & ~(~0 << pageshift)
        return (page, immediate)

    def write(self, reg, val, width=8):
        page, imm = self._snip_regaddr(reg)

        if self.paged and page != self.curr_page:
            self._switch_page(page)

        valbytes = val.to_bytes(width//8, byteorder="little")
        self.bus.write_reg(self.addr, imm, valbytes,
                            regaddrlen=self.regimmbytes)

    def read(self, reg, width=8):
        page, imm = self._snip_regaddr(reg)

        if self.paged and page != self.curr_page:
            self._switch_page(page)

        data = self.bus.read_reg(self.addr, imm, width//8,
                                    regaddrlen=self.regimmbytes)
        return int.from_bytes(data, byteorder='little')

    def __repr__(self):
        label = self.name or f"@ {self.addr:02x}"
        return f"<{type(self).__name__} {label}>"



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *

HV_VTMR_CTL = (3, 5, 15, 1, 3)
HV_VTMR_CTL_VMASK = (1 << 0)
HV_VTMR_CTL_PMASK = (1 << 1)

HV_VTMR_LIST = (3, 5, 15, 1, 2)

TGE = (1<<27)

u.msr(CNTHCTL_EL2, 3 << 10) # EL1PTEN | EL1PCTEN

def run_test(ctl, tval):
    u.inst(0xd5033fdf) # isb

    u.msr(ctl, 0)
    u.msr(tval, int(freq * 0.8))
    u.msr(ctl, 1)

    for i in range(6):
        p.nop()
        time.sleep(0.2)
        #u.inst(0xd5033fdf, call=p.el1_call)
        print("      . (ISR_EL1=%d) CTL=%x VTMR_LIST=%x" % (u.mrs(ISR_EL1), u.mrs(ctl), u.mrs(HV_VTMR_LIST)))

    u.msr(ctl, 0)

def test_hv_timers():
    u.msr(DAIF, 0x3c0)
    print("Testing HV timers...")
    print("  TGE = 1")

    u.msr(HCR_EL2, u.mrs(HCR_EL2) | TGE | (1 << 3) | (1 << 4))

    print("    P:")
    run_test(CNTP_CTL_EL0, CNTP_TVAL_EL0)
    print("    V:")
    run_test(CNTV_CTL_EL0, CNTV_TVAL_EL0)

def test_guest_timers():
    u.msr(DAIF, 0)
    print("Testing guest timers...")

    print("  TGE = 1, vGIC mode=0, timers unmasked")
    u.msr(HCR_EL2, (u.mrs(HCR_EL2) | TGE) | (1 << 3) | (1 << 4))
    u.msr(HACR_EL2, 0)
    u.msr(HV_VTMR_CTL, 3)

    print("    P:")
    #run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    #run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    print("  TGE = 1, vGIC mode=0, timers masked")
    u.msr(HV_VTMR_CTL, 0)

    print("    P:")
    run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    print("  TGE = 0, vGIC mode=0, timers unmasked")
    u.msr(HCR_EL2, (u.mrs(HCR_EL2) & ~TGE) | (1 << 3) | (1 << 4))
    u.msr(HACR_EL2, 0)
    u.msr(HV_VTMR_CTL, 3)

    print("    P:")
    run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    print("  TGE = 0, vGIC mode=0, timers masked")
    u.msr(HV_VTMR_CTL, 0)

    print("    P:")
    run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    print("  TGE = 0, vGIC mode=1, timers unmasked")
    u.msr(HCR_EL2, (u.mrs(HCR_EL2) & ~TGE) | (1 << 3) | (1 << 4))
    u.msr(HACR_EL2, 1<<20)
    u.msr(HV_VTMR_CTL, 3)

    print("    P:")
    run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    print("  TGE = 0, vGIC mode=1, timers masked")
    u.msr(HV_VTMR_CTL, 0)

    print("    P:")
    run_test(CNTP_CTL_EL02, CNTP_TVAL_EL02)
    print("    V:")
    run_test(CNTV_CTL_EL02, CNTV_TVAL_EL02)

    return

freq = u.mrs(CNTFRQ_EL0)
print("Timer freq: %d" % freq)

test_hv_timers()
test_guest_timers()
